"""
Created by adam on 3/14/20
"""
__author__ = 'adam'

from IPython.display import display
from ipywidgets import Layout, widgets, HBox, VBox

from CanvasHacks import environment
from CanvasHacks.Displays.dashboard import SkaaDashboard, DiscussionDashboard
from CanvasHacks.Repositories.overview import SkaaOverviewRepository, DiscussionOverviewRepository
from CanvasHacks.Widgets.AssignmentSelection import make_unit_chooser
from CanvasHacks.Widgets.DiscussionControls import discussion_run_button
# from CanvasHacks.Widgets.NagControls import nag_no_essay
from CanvasHacks.Widgets.LiveSelection import make_test_selector
from CanvasHacks.Widgets.NagControls import nag_button_area
from CanvasHacks.Widgets.SkaaControls import skaa_run_button

if __name__ == '__main__':
    pass


def dashboards_load_button( control_store, return_button=False, width='auto' ):
    """
    :param control_store:
    :param return_button: Set to true if putting inside a box or something else which will call display
    :return:
    """

    RUNNING = False

    def get_style( is_running=False ):
        return 'warning' if is_running else 'primary'

    def get_name( is_running=False ):
        return 'RUNNING' if is_running else 'REFRESH DASHBOARDS'

    # Create the button
    layout = widgets.Layout( width=width )
    b = widgets.Button( description=get_name( RUNNING ),
                        layout=layout,
                        button_style=get_style( RUNNING ) )

    def callback( change ):
        RUNNING = True
        b.description = get_name( RUNNING )
        b.button_style = get_style( RUNNING )

        control_store[ 'skaa_repo' ].load( environment.CONFIG.unit )
        control_store[ 'skaa_dash' ].print_counts()
        control_store[ 'diss_repo' ].load( environment.CONFIG.unit )
        control_store[ 'diss_dash' ].print_counts()

        RUNNING = False

        b.description = get_name( RUNNING )
        b.button_style = get_style( RUNNING )

    b.on_click( callback )
    if return_button is True:
        # If putting inside a box or something else which will
        # call display
        return b
    else:
        display( b )


def run_buttons( control_store, **kwargs ):
    box_layout = Layout(
        border='solid',
        **kwargs
        # width='50%'
    )

    box = widgets.VBox( [
        widgets.Label( value="Distribute work and results" ),
        skaa_run_button( control_store, return_button=True,  **kwargs),
        discussion_run_button( control_store, return_button=True, **kwargs )
    ], layout=box_layout )
    return box


def make_control_store():
    # Initialize these but do not load until a unit is selected
    sr = SkaaOverviewRepository()
    dr =  DiscussionOverviewRepository()
    return {
        'skaa_repo': sr,
        'diss_repo': dr,
        'skaa_dash' : SkaaDashboard(sr),
        'diss_dash' : DiscussionDashboard(dr)
    }


def button_area(control_store):
    """
    Displays all control buttons
    :param control_store:
    :return:
    """
    row_layout = Layout( width='100%', padding='10px' )

    row1 = HBox([make_test_selector(return_button=True)], layout=row_layout)

    row2 = VBox(make_unit_chooser(return_button=True, width='100%'), layout=row_layout)

    row3 = HBox([
        run_buttons(control_store=control_store, width='100%', padding='2px'),
        nag_button_area(control_store, width='100%', padding='2px')
    ], layout=row_layout)

    row4 = HBox(
        [
            dashboards_load_button(control_store, return_button=True, width='auto'),
        ],
        layout=row_layout)

    container = VBox(
        [row1, row2, row3, row4], layout=Layout(border='dashed', width='80%'))
    display(container)


"""
Created by adam on 2/24/20
"""
from CanvasHacks.Errors.data_ingestion import NoNewSubmissions

__author__ = 'adam'

if __name__ == '__main__':
    pass


class ILoaderFactory:

    @staticmethod
    def make( **kwargs ):
        raise NotImplementedError


class ILoader:
    """
    Defines interfaces for objects which load student
    data and return it as a dataframe.

    These shouldn't modify the data. That's the job of a
    repository.
    """

    @staticmethod
    def get_quiz( course, activity ):
        """Returns the canvasapi.quiz.Quiz object associated
        with this repository.
        """
        return course.get_quiz( activity.quiz_id )

    @staticmethod
    def get_assignment(course, activity):
        """If doesn't use quiz report, we need the activity_inviting_to_complete"""
        return course.get_assignment(activity.id)

    @staticmethod
    def load( activity, course=None, **kwargs ):
        raise NotImplementedError


class INewLoader(ILoader):
    """Interface for any class which ingests data and returns
    what hasn't been acted upon yet
    """

    @staticmethod
    def _check_empty( data ):
        """
        Should be called on what's been loaded before returning
        it.
        :raises NoNewSubmissions
        """
        if len( data ) == 0:
            raise NoNewSubmissions


class IAllLoader(ILoader):
    """Interface for any class which loads all existing
    data for the quiz"""

    @staticmethod
    def _check_empty( data ):
        """
        Should be called on what's been loaded before returning
        it.
        :raises NoNewSubmissions
        """
        if len( data ) == 0:
            raise NoNewSubmissions

    # @staticmethod
    # def load( activity_inviting_to_complete, course=None, **kwargs ):
    #     raise NotImplementedError


"""
Created by adam on 12/24/19
"""
from unittest import TestCase

from faker import Faker

from CanvasHacks.Models.student import Student, ensure_student
from tests.factories.ModelFactories import student_factory

fake = Faker()

__author__ = 'adam'

if __name__ == '__main__':
    pass


class TestStudentUtilities( TestCase ):

    def test_ensure_student_when_student( self ):
        s = student_factory()
        r = ensure_student( s )
        self.assertIsInstance( r, Student, "Returns student" )
        self.assertTrue( s == r )

    def test_ensure_student_int_id( self ):
        s = 1111114
        r = ensure_student( s )
        self.assertIsInstance( r, Student, "Returns student" )
        self.assertTrue( s == r.student_id )

    def test_ensure_student_string_id( self ):
        s = '1111114'
        r = ensure_student( s )
        self.assertIsInstance( r, Student, "Returns student" )
        self.assertTrue( int( s ) == r.student_id )

    def test_ensure_student_dict( self ):
        s = {'student_id' : '1111114',
             'other' : 'taco'}
        r = ensure_student( s )
        self.assertIsInstance( r, Student, "Returns student in dictionary case" )
        self.assertTrue( int( s['student_id'] ) == r.id )

    def test_ensure_student_obj( self ):
        self.skipTest('unsure if needed')
        # s = object()
        # setattr(s, 'student_id', '1111114')
        # r = ensure_student( s )
        # self.assertIsInstance( r, Student, "Returns student in object case" )
        # self.assertTrue( int( s['student_id'] ) == r.id )
        #


class TestStudent( TestCase ):

    # def test_handle_kwargs( self ):
    #     td = { 'student_id': fake.isbn10(),
    #            }
    #     kws = {
    #         'name': fake.name(),
    #         'taco': 'fish',
    #         'noses': 5 }
    #
    #     obj = Student( td[ 'student_id' ], **kws )
    #
    #     for k in kws.keys():
    #         self.assertEqual( getattr( obj, k ), kws[ k ] )

    def test_comparison_for_equality( self ):
        s1 = Student( student_id=333333 )
        s2 = Student( student_id=333333 )
        self.assertTrue( s1 == s2, "Correctly handles students w same id" )

        s1 = Student( student_id=333333 )
        s3 = Student( student_id=222222 )
        self.assertFalse( s1 == s3, "Correctly handles students w different ids" )



"""
Assigns a randomly selected reviewer and sends them the work

"""
from unittest.mock import MagicMock, patch

import CanvasHacks.testglobals
from CanvasHacks.Models.review_association import ReviewAssociation
from CanvasHacks.Repositories.status import InvitationStatusRepository
from tests.factories.ModelFactories import student_factory
from tests.factories.PeerReviewedFactories import unit_factory
from tests.factories.RepositoryMocks import ContentRepositoryMock

CanvasHacks.testglobals.use_api = False
from CanvasHacks.Errors.review_pairings import AllAssigned, NoAvailablePartner

from CanvasHacks.DAOs.sqlite_dao import SqliteDAO
from tests.TestingBase import TestingBase

from CanvasHacks.SkaaSteps.SendInitialWorkToReviewer import SendInitialWorkToReviewer
from faker import Faker

fake = Faker()


class TestCallsAllExpected( TestingBase ):
    """Makes sure that everything gets called with expected values.
      Not super diagnostic since many of the mocked calls are
      where we'd actually expect failure. Still, useful for catching
      problems when update code etc"""

    def setUp( self ):
        self.config_for_test()
        self.dao = SqliteDAO()

        self.course = MagicMock()
        # review = Review(**activity_data_factory())
        self.unit = unit_factory()
        # self.unit.components.append(review)

    @patch( 'CanvasHacks.SkaaSteps.SendInitialWorkToReviewer.InvitationStatusRepository' )
    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendInitialWorkToReviewer.PeerReviewInvitationMessenger' )
    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.AssociationRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendInitialWorkToReviewer.WorkRepositoryLoaderFactory' )
    def test_run( self, workLoaderMock, assocRepoMock, messengerMock, studentRepoMock, statusRepoMock ):
        """Check that each student receives the expected message
        containing the correct student's submission
        """
        students = [ student_factory(), student_factory() ]

        def student_getter( sid ):
            return [ s for s in students if s.student_id == sid ][ 0 ]

        submitter_ids = [ s.student_id for s in students ]
        workRepo = MagicMock()
        workRepo.submitter_ids = submitter_ids
        workLoaderMock.make = MagicMock( return_value=workRepo )

        studentRepoMock.download = MagicMock( return_value=students )
        studentRepoMock.get_student = student_getter

        obj = SendInitialWorkToReviewer( course=self.course, unit=self.unit, is_test=True, send=True )

        assignments = [ ReviewAssociation( assessee_id=students[ 0 ], assessor_id=students[ 1 ] ),
                        ReviewAssociation( assessee_id=students[ 1 ], assessor_id=students[ 0 ] ), ]
        obj.associationRepo.assign_reviewers = MagicMock( return_value=assignments )

        # call
        obj.run()

        # check
        workLoaderMock.make.assert_called()
        workLoaderMock.make.assert_called_with( self.unit.initial_work, self.course) # only_new=False, rest_timeout=5 )

        obj.studentRepo.download.assert_called()

        obj.associationRepo.assign_reviewers.assert_called()
        obj.associationRepo.assign_reviewers.assert_called_with( submitter_ids )

        obj.messenger.notify.assert_called()
        obj.messenger.notify.assert_called_with( assignments, True )
        # obj.messenger.notify.assert_called_with( obj.associationRepo.data, True )

        # obj.statusRepo.record_invited.assert_called()
        # obj.statusRepo.record_invited.assert_called_with(submitter_ids[0])
        # obj.statusRepo.record_invited.assert_called_with(submitter_ids[1])


class TestUnitTests(TestingBase):
    """Unit tests with all dependencies mocked"""
    def setUp( self ):
        self.config_for_test()
        self.unit = unit_factory()
        self.course = MagicMock()
        self.activity_id = self.unit.review.id
        # self.dao = SqliteDAO()
        # self.create_new_and_preexisting_students()
        # # Prepare fake work repo to give values to calling  objects
        # self.workRepo = ContentRepositoryMock()
        # self.workRepo.create_test_content( self.student_ids )
        # self.workRepo.add_students_to_data(self.student_ids, make_dataframe=True)

    def test_instantiates_correct_status_repos( self ):
        """The sending of metareview results requires a
        special status repository"""
        obj = SendInitialWorkToReviewer( course=self.course, unit=self.unit, is_test=True, send=True )

        self.assertIsInstance(obj.statusRepos, list, "status repos is a list")
        self.assertTrue(len(obj.statusRepos) == 1)
        self.assertIsInstance(obj.invite_status_repo, InvitationStatusRepository, "Invite repo instantiated")



class TestFunctionalTestWhenQuizType( TestingBase ):
    """
    Tests the entire process in various use cases using
    local data when the unit is some flavor of quiz (and thus
    uses a quiz report)

    For tests of whole process which hit server, see tests.EndToEnd
    """

    def setUp( self ):
        self.config_for_test()
        self.unit = unit_factory()
        self.course = MagicMock()

        self.activity_id = self.unit.initial_work.id
        # self.dao = SqliteDAO()
        # self.session = self.dao.session
        self.create_new_and_preexisting_students()
        # Prepare fake work repo to give values to calling  objects
        self.workRepo = ContentRepositoryMock()
        self.workRepo.create_test_content( self.student_ids )
        self.workRepo.add_students_to_data(self.student_ids, make_dataframe=True)

    @patch( 'CanvasHacks.SkaaSteps.SendInitialWorkToReviewer.InvitationStatusRepository' )
    @patch( 'CanvasHacks.Messaging.base.ConversationMessageSender.send' )
    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendInitialWorkToReviewer.WorkRepositoryLoaderFactory' )
    def test_does_not_send_to_people_already_assigned( self, workLoaderMock, studentRepoMock, messengerMock,
                                                       statusRepoMock ):
        """
        Simulates a run where some students have already been assigned
        peer reviewers and a new group is being processed.

        Focus in testing is on making sure that the two groups don't get mixed up

        :param workLoaderMock:
        :param studentRepoMock:
        :param messengerMock:
        :return:
        """

        # Prepare fake work repo to give values to calling  objects
        # workRepo = ContentRepositoryMock()
        # workRepo.create_test_content( self.student_ids )
        # workRepo = create_autospec( QuizRepository )
        # workRepo.get_formatted_work_by = MagicMock( return_value=testText )
        self.workRepo.submitter_ids = self.new_students_ids
        workLoaderMock.make = MagicMock( return_value=self.workRepo )

        # prepare student repo
        students = { s.student_id: s for s in self.students }

        def se( sid ):
            return students.get( sid )

        # studentRepoMock.get_student = MagicMock(side_effect=se)
        # studentRepoMock.download = MagicMock( return_value=self.students )
        # feedbackStatusRepoMock.record_invited = MagicMock()
        # call
        obj = SendInitialWorkToReviewer( course=self.course, unit=self.unit, is_test=True, send=True )
        # obj.studentRepo = MagicMock()
        obj.studentRepo.get_student = MagicMock( side_effect=se )
        obj.studentRepo.download = MagicMock( return_value=self.students )

        # Have to do this after object creation so that we can use the
        # same in-memory db
        self.session = obj.dao.session
        preexisting_pairings = self.create_preexisting_review_pairings( self.activity_id, self.preexisting_students )

        obj.run()

        # check
        self.assertEqual( len( obj.associations ), len( self.new_students_ids ),
                          "Correct number of students assigned" )

        # Check that new assignments don't involve previously assigned sudents
        for rec in obj.associations:
            self.assertNotIn( rec.assessor_id, self.preexisting_student_ids,
                              "Newly assigned assessor not among previously assigned students" )
            self.assertNotIn( rec.assessee_id, self.preexisting_student_ids,
                              "Newly assigned assessee not among previously assigned students" )

        # Check whether each new student has been assigned
        new_assessor_ids = [ r.assessor_id for r in obj.associations ]
        new_assessee_ids = [ r.assessee_id for r in obj.associations ]
        self.assertEqual( len( set( new_assessor_ids ) ), len( self.new_students_ids ),
                          "No duplicate assessor assignments" )
        self.assertEqual( len( set( new_assessee_ids ) ), len( self.new_students_ids ),
                          "No duplicate assessee assignments" )
        for sid in self.new_students_ids:
            self.assertIn( sid, new_assessee_ids, "Student is an assessee" )
            self.assertIn( sid, new_assessor_ids, "Student is an assessor" )

        # ================== Events on Messenger
        # Check that mocked objects were called with expected data
        messengerMock.assert_called()
        self.assertEqual( messengerMock.call_count, len( self.new_students_ids ),
                          "Send method called expected number of times" )
        messenger_args = [ (c[ 1 ][ 'student_id' ], c[ 1 ][ 'subject' ], c[ 1 ][ 'body' ]) for c in
                           messengerMock.call_args_list ]
        # print(args)

        # Status repo calls on messenger
        # feedbackStatusRepoMock.record_invited.assert_called()
        obj.messenger.status_repositories[0].record.assert_called()
        call_list = obj.messenger.status_repositories[0].record.call_args_list
        status_args = [ c[ 0 ][ 0 ] for c in call_list ]
        self.assertEqual( len( self.new_students ), len( call_list ),
                          "Status repo record_invited called expected number of times" )
        for sid in self.new_students_ids:
            self.assertIn( sid, status_args, "StatusRepo.record called on all students" )

        # student repo calls on messenger
        for sid in self.new_students_ids:
            obj.messenger.student_repository.get_student.assert_any_call( sid )

        # Check the content sent
        for record in obj.associations:
            # print(record.assessee_id)
            author_text = self.workRepo.get_formatted_work_by( record.assessee_id )
            # see if sent to assessor
            sent_text = [ t[ 2 ] for t in messenger_args if t[ 0 ] == record.assessor_id ][ 0 ]
            rx = r'{}'.format( author_text )
            self.assertRegex( sent_text, rx, "Author's work in message sent to reviewer" )

    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendInitialWorkToReviewer.WorkRepositoryLoaderFactory' )
    def test_raises_when_all_submitters_already_assigned( self, workLoaderMock, studentRepoMock ):

        # Prepare fake work repo to give values to calling  objects
        # workRepo = ContentRepositoryMock()
        # workRepo.create_test_content( self.preexisting_student_ids )
        # workRepo = create_autospec( QuizRepository )
        # workRepo.get_formatted_work_by = MagicMock( return_value=testText )

        # Setting to return all the previously assigned students.
        # this should cause all the students to be filtered out
        # and the errror raised
        self.workRepo.submitter_ids = self.preexisting_student_ids
        workLoaderMock.make = MagicMock( return_value=self.workRepo )

        studentRepoMock.download = MagicMock( return_value=self.students )

        # call
        obj = SendInitialWorkToReviewer( course=self.course, unit=self.unit, is_test=True, send=True )
        self.session = obj.dao.session
        preexisting_pairings = self.create_preexisting_review_pairings( self.activity_id, self.preexisting_students )

        with self.assertRaises( AllAssigned ):
            obj.run()

    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendInitialWorkToReviewer.WorkRepositoryLoaderFactory' )
    def test_raises_when_only_one_submission( self, workLoaderMock, studentRepoMock ):

        # Prepare fake work repo to give values to calling  objects
        # workRepo = ContentRepositoryMock()
        # workRepo.create_test_content( self.preexisting_student_ids )
        # workRepo = create_autospec( QuizRepository )
        # workRepo.get_formatted_work_by = MagicMock( return_value=testText )

        # Setting to return all the previously assigned students.
        # this should cause all the students to be filtered out
        # and the errror raised
        self.workRepo.submitter_ids = self.preexisting_student_ids
        workLoaderMock.make = MagicMock( return_value=self.workRepo )

        studentRepoMock.download = MagicMock( return_value=self.students )

        # call
        obj = SendInitialWorkToReviewer( course=self.course, unit=self.unit, is_test=True, send=True )
        # Have to do this after object creation so that we can use the
        # same in-memory db
        self.session = obj.dao.session
        preexisting_pairings = self.create_preexisting_review_pairings( self.activity_id, self.preexisting_students )

        with self.assertRaises( AllAssigned ):
            obj.run()


class TestFunctionalTestWhenNonQuizType( TestingBase ):
    """
    Tests the entire process in various use cases using
    local data when the unit is NOT some flavor of quiz (and thus
    does not use a quiz report)

    For tests of whole process which hit server, see tests.EndToEnd
    """

    def setUp( self ):
        self.config_for_test()
        self.unit = unit_factory()
        self.course = MagicMock()

        # Set to none so that loader thinks not a quiz
        self.unit.initial_work.quiz_id = None
        self.activity_id = self.unit.initial_work.id
        self.dao = SqliteDAO()
        self.session = self.dao.session
        self.create_new_and_preexisting_students()
        # Prepare fake work repo to give values to calling  objects
        self.workRepo = ContentRepositoryMock()
        self.workRepo.create_test_content( self.student_ids )
        self.workRepo.add_students_to_data(self.student_ids, make_dataframe=True)

        # workRepo = create_autospec( QuizRepository )
        # workRepo.get_formatted_work_by = MagicMock( return_value=testText )

    @patch( 'CanvasHacks.SkaaSteps.SendInitialWorkToReviewer.InvitationStatusRepository' )
    @patch( 'CanvasHacks.Messaging.base.ConversationMessageSender.send' )
    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendInitialWorkToReviewer.WorkRepositoryLoaderFactory' )
    def test_does_not_send_to_people_already_assigned( self, workLoaderMock, studentRepoMock, messengerMock,
                                                       statusRepoMock ):
        """
        Simulates a run where some students have already been assigned
        peer reviewers and a new group is being processed.

        Focus in testing is on making sure that the two groups don't get mixed up

        :param workLoaderMock:
        :param studentRepoMock:
        :param messengerMock:
        :return:
        """
        self.workRepo.submitter_ids = self.new_students_ids

        workLoaderMock.make = MagicMock( return_value=self.workRepo )

        # prepare student repo
        students = { s.student_id: s for s in self.students }

        def se( sid ):
            return students.get( sid )

        # studentRepoMock.get_student = MagicMock(side_effect=se)
        # studentRepoMock.download = MagicMock( return_value=self.students )

        # call
        obj = SendInitialWorkToReviewer( course=self.course, unit=self.unit, is_test=True, send=True )
        # obj.studentRepo = MagicMock()
        obj.studentRepo.get_student = MagicMock( side_effect=se )
        obj.studentRepo.download = MagicMock( return_value=self.students )
        preexisting_pairings = self.create_preexisting_review_pairings( self.activity_id, self.preexisting_students,
                                                                        obj.dao.session )

        obj.run()

        # check
        self.assertEqual( len( obj.associations ), len( self.new_students_ids ),
                          "Correct number of students assigned" )

        # Check that new assignments don't involve previously assigned sudents
        for rec in obj.associations:
            self.assertNotIn( rec.assessor_id, self.preexisting_student_ids,
                              "Newly assigned assessor not among previously assigned students" )
            self.assertNotIn( rec.assessee_id, self.preexisting_student_ids,
                              "Newly assigned assessee not among previously assigned students" )

        # Check whether each new student has been assigned
        new_assessor_ids = [ r.assessor_id for r in obj.associations ]
        new_assessee_ids = [ r.assessee_id for r in obj.associations ]
        self.assertEqual( len( set( new_assessor_ids ) ), len( self.new_students_ids ),
                          "No duplicate assessor assignments" )
        self.assertEqual( len( set( new_assessee_ids ) ), len( self.new_students_ids ),
                          "No duplicate assessee assignments" )
        for sid in self.new_students_ids:
            self.assertIn( sid, new_assessee_ids, "Student is an assessee" )
            self.assertIn( sid, new_assessor_ids, "Student is an assessor" )

        # ================== Events on Messenger
        # Check that mocked objects were called with expected data
        messengerMock.assert_called()
        self.assertEqual( messengerMock.call_count, len( self.new_students_ids ),
                          "Send method called expected number of times" )
        messenger_args = [ (c[ 1 ][ 'student_id' ], c[ 1 ][ 'subject' ], c[ 1 ][ 'body' ]) for c in
                           messengerMock.call_args_list ]
        # print(args)

        # Check that all messages have the correct subject
        for sid, subj, body in messenger_args:
            self.assertEqual(self.unit.review.email_subject, subj, "Correct subject line")

        # Status repo calls on messenger (only 1 repo)
        obj.messenger.status_repositories[0].record.assert_called()
        call_list = obj.messenger.status_repositories[0].record.call_args_list

        # obj.messenger.status_repositories.record_invited.assert_called()
        # call_list = obj.messenger.status_repositories.record_invited.call_args_list
        status_args = [ c[ 0 ][ 0 ] for c in call_list ]
        self.assertEqual( len( self.new_students ), len( call_list ),
                          "Status repo record_invited called expected number of times" )
        for sid in self.new_students_ids:
            self.assertIn( sid, status_args, "StatusRepo.record_invited called on all students" )

        # student repo calls on messenger
        for sid in self.new_students_ids:
            obj.messenger.student_repository.get_student.assert_any_call( sid )

        # Check the content sent
        for record in obj.associations:
            # print(record.assessee_id)
            author_text = self.workRepo.get_formatted_work_by( record.assessee_id )
            # see if sent to assessor
            sent_text = [ t[ 2 ] for t in messenger_args if t[ 0 ] == record.assessor_id ][ 0 ]
            rx = r'{}'.format( author_text )
            self.assertRegex( sent_text, rx, "Author's work in message sent to reviewer" )

    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendInitialWorkToReviewer.WorkRepositoryLoaderFactory' )
    def test_raises_when_all_submitters_already_assigned( self, workLoaderMock, studentRepoMock ):
        # Prepare fake work repo to give values to calling  objects
        # workRepo = ContentRepositoryMock()
        # workRepo.create_test_content( self.preexisting_student_ids )
        # workRepo.add_students_to_data(self.student_ids, make_dataframe=True)

        # Setting to return all the previously assigned students.
        # this should cause all the students to be filtered out
        # and the errror raised
        self.workRepo.submitter_ids = self.preexisting_student_ids
        workLoaderMock.make = MagicMock( return_value=self.workRepo )

        studentRepoMock.download = MagicMock( return_value=self.students )

        # call
        obj = SendInitialWorkToReviewer( course=self.course, unit=self.unit, is_test=True, send=True )
        self.session = obj.dao.session
        preexisting_pairings = self.create_preexisting_review_pairings( self.activity_id, self.preexisting_students )

        with self.assertRaises( AllAssigned ):
            obj.run()

    @patch( 'CanvasHacks.SkaaSteps.ISkaaSteps.StudentRepository' )
    @patch( 'CanvasHacks.SkaaSteps.SendInitialWorkToReviewer.WorkRepositoryLoaderFactory' )
    def test_raises_when_only_one_submission( self, workLoaderMock, studentRepoMock ):

        # Prepare fake work repo to give values to calling  objects
        workRepo = ContentRepositoryMock()
        workRepo.create_test_content( self.new_students_ids )
        workRepo.add_students_to_data(self.student_ids, make_dataframe=True)

        # Setting to return all the previously assigned students.
        # this should cause all the students to be filtered out
        # and the errror raised
        self.workRepo.submitter_ids = [ self.new_students[ 0 ] ]
        workLoaderMock.make = MagicMock( return_value=self.workRepo )

        studentRepoMock.download = MagicMock( return_value=self.students )

        # call
        obj = SendInitialWorkToReviewer( course=self.course, unit=self.unit, is_test=True, send=True )
        # Have to do this after object creation so that we can use the
        # same in-memory db
        self.session = obj.dao.session
        preexisting_pairings = self.create_preexisting_review_pairings( self.activity_id, self.preexisting_students )

        with self.assertRaises( NoAvailablePartner ):
            obj.run()



"""
Created by adam on 10/1/19
"""
__author__ = 'adam'

from CanvasHacks.Models.model import StoreMixin


class IRepo( StoreMixin ):
    """Parent class for all repositories.
    All will store data in the data attribute,
    but this does not constrain how the data is stored.
    Some use a dataframe, others use a dictionary
    """
    pass


class IDownloadingRepo:
    """Repositories which download stuff"""

    def download( self ):
        raise NotImplementedError


class IContentRepository( IRepo ):
    """Defines methods of repositories which
    return the content of student work
    """


    def get_formatted_work_by( self, student_id ):
        raise NotImplementedError


class ISubmissionRepo( IRepo, IDownloadingRepo ):

    def download( self ):
        raise NotImplementedError

    @property
    def frame( self ):
        raise NotImplementedError

    def get_by_student_id( self, student_id ):
        raise NotImplementedError


if __name__ == '__main__':
    pass


"""
Created by adam on 1/7/19
"""
__author__ = 'adam'

from IPython.display import display
from ipywidgets import widgets

from CanvasHacks.Configuration import InteractiveConfiguration


# -------- General
def make_text_input( input_dict ):
    """Creates a text input field. The given dictionary should have keys 'label' and 'handler'"""
    text = widgets.Text( description=input_dict[ 'label' ] )
    display( text )
    text.observe( input_dict[ 'handler' ] )
    return text


# -------- Canvas specific
def make_canvas_token_input():
    """Creates the input field for the canvas api token"""
    canvas_token_input = { 'label': 'Canvas token', 'handler': InteractiveConfiguration.handle_token_entry }
    return make_text_input( canvas_token_input )


def make_canvas_url_input():
    """Creates the input field for the canvas api token"""
    canvas_url_input = { 'label': 'Canvas url', 'handler': InteractiveConfiguration.handle_url_entry }
    make_text_input( canvas_url_input )


def make_course_ids_input():
    """
    Make box and buttons for course id entry.
    On submit button click, class ids updates with the value in the text box
    """
    t = widgets.Text( description='Class id' )
    submit = widgets.Button( description='Add id', button_style='success' )
    reset = widgets.Button( description='Reset ids' )
    out = widgets.Output( layout={ 'border': '1px solid black' } )

    def handle_add( change ):
        InteractiveConfiguration.add_course_id( t.value )
        # Clear the box
        t.value = ''
        show_class_ids()

    def handle_reset( change ):
        InteractiveConfiguration.reset_course_ids()
        show_class_ids()

    def show_class_ids():
        out.clear_output()
        with out:
            print( '    ----Canvas class ids----    ' )
            for cid in InteractiveConfiguration.course_ids:
                print( cid )

    submit.on_click( handle_add )
    reset.on_click( handle_reset )
    button_box = widgets.HBox( [ reset, submit ] )
    input_box = widgets.VBox( [ t, button_box ] )
    display( widgets.HBox( [ input_box, out ] ) )


def make_general_reset_button():
    rb = widgets.Button( description='Clear canvas token and class ids' )

    def reset_all( change ):
        InteractiveConfiguration.reset_config()

    display( rb.on_click( reset_all ) )


if __name__ == '__main__':
    pass



"""
Created by adam on 2/23/20
"""
__author__ = 'adam'

from sqlalchemy import Column, Integer
from sqlalchemy.dialects.sqlite import DATETIME

from CanvasHacks import environment as env
# from sqlalchemy.ext.declarative import declarative_base
from CanvasHacks.DAOs.sqlite_dao import Base
from CanvasHacks.Models.model import Model
from CanvasHacks.TimeTools import current_utc_timestamp

# Deprecated
class StatusRecord( Base, Model ):
    """
    :deprecated

    Keeps track of when an activity_inviting_to_complete was made available to a
    student and when they submitted it.

    NB, No need for recording when a rest_timeout notification is sent.
    That can go in the logs for auditing
    """
    __tablename__ = env.STATUS_TABLE_NAME

    # Id of the relevant activity_inviting_to_complete
    activity_id = Column(Integer, primary_key=True, nullable=False)

    # The student whom this record concerns
    student_id = Column( Integer, primary_key=True, nullable=False )

    # When this student submitted the activity_inviting_to_complete
    submitted = Column( DATETIME, nullable=True )

    # When the student was notified this was available
    # e.g., for metareview, this would be immediately after review submitted
    # by other party
    notified = Column( DATETIME, nullable=True )

    # When they were sent metareview feedback was sent out
    #
    # This is only relevant for the metareview since need to record when the
    # feedback from this student was sent out to the person who did the peer review.
    # Notified already represents when they were invited to do the metareview unit
    results = Column(DATETIME, nullable=True)

    def record_submission( self, time_to_record=None ):
        """
        Called to record the timestamp of when this student submitted
        their content unit
        :param time_to_record:
        :return:
        """
        if time_to_record is None:
            time_to_record = current_utc_timestamp()
        self.submitted = time_to_record

    def record_opened( self, time_to_record=None ):
        """
        Called to record the timestamp of when this student was
        notified that the activity_inviting_to_complete was available.
        With activities like peer review, this is also the
        time that they were sent someone else's work
        :param time_to_record:
        :return:
        """
        if time_to_record is None:
            time_to_record = current_utc_timestamp()
        self.notified = time_to_record

    def record_sent_results( self, time_to_record=None ):
        """
        Called to record the timestamp of when they were sent
        feedback on the unit
        :param time_to_record:
        :return:
        """

        if time_to_record is None:
            time_to_record = current_utc_timestamp()
        self.results = time_to_record



class FeedbackReceivedRecord( Base, Model ):
    """
    Keeps track of when a student was sent feedback from
    another student

    Created in CAN-53

    """
    __tablename__ = env.FEEDBACK_RECEIVED_STATUS_TABLE_NAME

    activity_id = Column(Integer, primary_key=True, nullable=False)
    """Id of the activity in which the feedback was created"""

    student_id = Column( Integer, primary_key=True, nullable=False )
    """ The student whom this record concerns """

    sent_at = Column(DATETIME, nullable=True)
    """ Timestamp of when the feedback was sent to them"""



class InvitationReceivedRecord( Base, Model ):
    """
    Keeps track of when a student was sent an invitation to
     provide feedback on another student

    Created in CAN-53

    """
    __tablename__ = env.INVITATION_RECEIVED_STATUS_TABLE_NAME

    # Id of the relevant activity_inviting_to_complete
    activity_id = Column(Integer, primary_key=True, nullable=False)
    """Id of the activity which we are inviting the student to complete"""

    # The student whom this record concerns
    student_id = Column( Integer, primary_key=True, nullable=False )
    """ The student whom this record concerns """

    sent_at = Column(DATETIME, nullable=True)
    """ Timestamp of when the feedback was sent to them"""












class ComplexStatusRecord( Base, Model ):
    __tablename__ = env.COMPLEX_STATUS_TABLE_NAME
    # The student whom this record concerns
    student_id = Column( Integer, primary_key=True, nullable=False )

    # The id of the content unit
    content_assignment_id = Column( Integer, primary_key=True, nullable=False )

    # Id of student being reviewed by this student
    reviewer_of = Column( Integer, nullable=True )

    # Id of student reviewing this student
    reviewed_by = Column( Integer, nullable=True )

    # ------------------- Dates that this student did things
    # When this student submitted their content unit
    content_assignment_submitted = Column( DATETIME, nullable=True )

    # When this student submitted their review of the reviewer_of student
    review_submitted = Column( DATETIME, nullable=True )

    # When this student turned in their review of the feedback given by
    # the reviewed_by student
    metareview_submitted = Column( DATETIME, nullable=True )

    # ------------------- Dates when this student was sent things
    # Date that the person who reviews this student was sent the
    # reviewer_of student's content unit to review
    reviewer_assigned_on = Column( DATETIME, nullable=True )

    # If the student was sent a message that they must rest_timeout
    # until someone else submits the unit before they can review
    # this is when that notificaion was sent
    wait_notification_on = Column( DATETIME, nullable=True )

    # When this student was sent the results of the metareview by
    # the reviewer_of student
    metareview_results_on = Column( DATETIME, nullable=True )

    # When this student was sent the feedback created by
    # the reviewed_by student
    review_results_on = Column( DATETIME, nullable=True )

    def is_under_review( self ):
        """Returns true if someone has been assigned to
        review this students' content unit, false otherwise
        """
        return self.reviewed_by is not None

    def add_reviewee( self, reviewee, assigned_time=None ):
        """
        Sets the provided student as the person to be reviewed by
        the user whom the record belongs to.
        If assigned_time is set, it will use that as the reviewer_assigned_on value
        Otherwise will use current utc timestamp
        :param reviewee: Student object or id of student to be reviewed
        :param assigned_time:
        :return:
        """
        try:
            reviewee_id = reviewee.id
        except AttributeError:
            reviewee_id = reviewee

        if assigned_time is None:
            assigned_time = current_utc_timestamp()

        self.reviewer_assigned_on = assigned_time
        self.reviewer_of = reviewee_id

    def add_reviewer( self, reviewer ):
        """
        Sets the provided student as the person who is reviewing the
        the user whom the record belongs to.
        :param reviewer: Student object or id of student doing reviewing
        :return:
        """
        try:
            reviewer_id = reviewer.id
        except AttributeError:
            reviewer_id = reviewer

        self.reviewed_by = reviewer_id

    def record_content_assignment_submission( self, submitted_time=None ):
        """
        Called to record the timestamp of when this student submitted
        their content unit
        :param submitted_time:
        :return:
        """
        if submitted_time is None:
            submitted_time = current_utc_timestamp()
        self.content_assignment_submitted = submitted_time

    def record_sending_metareview_results( self, sent_time=None ):
        """Called to record the feedback from the reviewer_of student being
        sent to this student"""
        if sent_time is None:
            sent_time = current_utc_timestamp()
        self.metareview_results_on = sent_time

    def record_sending_review_results( self, sent_time=None ):
        """Call to record that the student has been sent the content submitted
        by the reviewed_by student"""
        if sent_time is None:
            sent_time = current_utc_timestamp()
        self.review_results_on = sent_time

    def record_wait_notification( self, sent_time=None ):
        """Call to record that the student has been sent a notification that
        they will have to rest_timeout to receive student work for them to review"""
        if sent_time is None:
            sent_time = current_utc_timestamp()
        self.wait_notification_on = sent_time


if __name__ == '__main__':
    pass



"""
Created by adam on 3/2/20
"""
__author__ = 'adam'

from CanvasHacks.Messaging.discussions import DiscussionReviewInvitationMessenger
from CanvasHacks.Repositories.status import InvitationStatusRepository

"""
Created by adam on 2/23/20
"""
import CanvasHacks.testglobals
from CanvasHacks.Errors.data_ingestion import NoNewSubmissions
from CanvasHacks.Errors.review_pairings import AllAssigned, NoAvailablePartner
from CanvasHacks.Logging.run_data import RunLogger
# from CanvasHacks.Repositories.factories import WorkRepositoryLoaderFactory
from CanvasHacks.Logging.review_pairings import make_review_audit_file
from CanvasHacks.SkaaSteps.ISkaaSteps import IStep

__author__ = 'adam'

from CanvasHacks.Repositories.factories import WorkRepositoryLoaderFactory


class SendForumPostsToReviewer( IStep ):
    # post_threshold: int

    def __init__( self, course=None, unit=None, is_test=None, send=True, post_threshold=None, **kwargs ):
        """
        :param course:
        :param unit:
        :param is_test:
        :param send: Whether to actually send the messages
        """
        super().__init__( course, unit, is_test, send, **kwargs )

        # The activity whose results we are going to be doing something with
        self.post_threshold = post_threshold
        self.activity = unit.discussion_forum

        # The activity which we are inviting the receiving student to complete
        self.activity_notifying_about = unit.discussion_review

        # The activity_inviting_to_complete whose id is used to store review pairings for the whole SKAA
        # We use the discussion review so that the invite status records
        # and feedback status records will have the same id, which makes it
        # easier to handle
        self.activity_for_review_pairings = unit.discussion_review

        self._initialize()

        # Initialize the relevant status repo

        self.invite_status_repo = InvitationStatusRepository( self.dao, self.activity_notifying_about )
        self.statusRepos = [self.invite_status_repo]

        # self.statusRepos = StatusRepository( self.dao, self.activity_notifying_about )

    def run( self, **kwargs ):
        """
        Loads discussion forum posts, assigns reviewers, and sends formatted
        work to reviewer
         only_new=False, rest_timeout=5
        :param rest_timeout: Number of seconds to rest_timeout for canvas to generate report
        :param only_new: Probably will not be used
        :return:
        """
        try:
            self._load_step( **kwargs )

            self._assign_step()

            self._message_step()

        except NoNewSubmissions:
            # Check if new submitters, bail if not
            print( "No new submissions" )
            # todo Log run failure
            RunLogger.log_no_submissions( self.activity )
            if CanvasHacks.testglobals.TEST:
                # Reraise so can see what happened for tests
                raise NoNewSubmissions

        except AllAssigned as e:
            # Folks already assigned to review
            # have somehow gotten past the new submissions filter
            # This could be due to them resubmitting late
            print( "New submitters but everyone already assigned" )
            # print( e.submitters )
            if CanvasHacks.testglobals.TEST:
                # Reraise so can see what happened for tests
                raise AllAssigned( e.submitters )

        except NoAvailablePartner as e:
            # We will probably want to notify the student that they
            # have had their work noticed, but they are now waiting for
            # a partner
            # todo Notify student that they are waiting
            print( "1 student has submitted and has no partner ") #, e.submitters )
            if CanvasHacks.testglobals.TEST:
                # Reraise so can see what happened for tests
                raise NoAvailablePartner( e.submitters )

    def _message_step( self ):
        # Send the work to the reviewers
        # Note that we still do this even if send is false because
        # the messenger will print out the messages rather than sending them
        self.messenger = DiscussionReviewInvitationMessenger( self.unit,
                                                              self.studentRepo,
                                                              self.work_repo,
                                                              self.statusRepos )

        # NB, we don't use associationRepo.data because we only
        # want to send to people who are newly assigned
        self.messenger.notify( self.associations, self.send )

        #  todo Want some way of tracking if messages fail to send so can resend
        # Log the run
        msg = "New peer review assignments: \t {} \n Notices sent successfully: \t {} \n Send errors: \t {} \n Errors: {}".format(
            len( self.associations ), self.messenger.sent_count, len( self.messenger.send_errors ),
            self.messenger.send_errors )  # self.associations )
        print( msg )
        RunLogger.log_reviews_assigned( self.activity_notifying_about, msg )

    def _assign_step( self ):
        # Assign reviewers to each submitter and store in db
        # NB, the repository will take care of removing already assigned
        # reviewers, so we just give them all submitters
        self.associations = self.associationRepo.assign_reviewers( self.work_repo.submitter_ids )
        if not self.is_test:
            # Save a more readable copy of all the assignments
            # to file
            make_review_audit_file( self.associationRepo, self.studentRepo, self.unit )

    def _load_step( self, **kwargs ):
        self.work_repo = WorkRepositoryLoaderFactory.make( self.activity, self.course, **kwargs )
        if self.post_threshold is not None:
            # If a minimum post count has been set, we toss out students
            # who have not reached that count
            self.work_repo.data = self.work_repo.filter_by_count(self.post_threshold)

        self.display_manager.initially_loaded = self.work_repo.data

    @property
    def audit_frame( self ):
        return self.associationRepo.audit_frame( self.studentRepo )



if __name__ == '__main__':
    pass



"""
Abstract classes to use for avoiding circular imports

Created by adam on 3/17/20
"""
__author__ = 'adam'

if __name__ == '__main__':
    pass



class DiscussionType:
    pass


class BlockableActivity:
    """Defines the activity as one whose completion
    can be blocked by another student's inaction
    """
    pass


"""
Created by adam on 3/2/20
"""
__author__ = 'adam'

from unittest.mock import MagicMock, create_autospec, patch

from CanvasHacks.Messaging.discussions import DiscussionReviewInvitationMessenger
from CanvasHacks.Repositories.status import StatusRepository
from CanvasHacks.Repositories.students import StudentRepository
from factories.ModelFactories import student_factory
from factories.PeerReviewedFactories import unit_factory
from factories.RepositoryMocks import ContentRepositoryMock
from tests.TestingBase import TestingBase

if __name__ == '__main__':
    pass
from CanvasHacks.Messaging.templates import DISCUSSION_REVIEW_NOTICE_TEMPLATE
from faker import Faker

fake = Faker()


class TestDiscussionReviewInvitationMessenger( TestingBase ):

    def setUp( self ):
        self.config_for_test()
        self.unit = unit_factory()

        self.activity = self.unit.discussion_review

        # student recieiving the message
        self.author = student_factory()
        self.reviewer = student_factory()

        # This would be metareview feedback on the review
        self.work = fake.text()

        self.studentRepo = StudentRepository()
        self.studentRepo.get_student = MagicMock( return_value=self.reviewer )
        self.contentRepo = ContentRepositoryMock()
        self.contentRepo.get_formatted_work_by = MagicMock( return_value=self.work )

        self.review_assign = MagicMock( assessor_id=self.reviewer.id, assessee_id=self.author.id )
        self.statusRepo = create_autospec( StatusRepository )

    def test_prepare_message( self ):
        obj = DiscussionReviewInvitationMessenger( self.unit, self.studentRepo, self.contentRepo, self.statusRepo )

        # call
        message_data = obj.prepare_message( self.review_assign )

        # check
        self.assertEqual( obj.message_template, DISCUSSION_REVIEW_NOTICE_TEMPLATE, "Working off expected template" )
        self.assertEqual( message_data[ 'student_id' ], self.reviewer.id, "Message is going to reviewer" )
        self.assertEqual( message_data[ 'subject' ], self.activity.email_subject, "Email subject" )

        self.assertTrue( len( message_data[ 'body' ] ) > 0 )

        d = obj._make_template_input( self.work, None, self.reviewer )

        expected_content = DISCUSSION_REVIEW_NOTICE_TEMPLATE.format( **d )

        self.assertEqual( expected_content, message_data[ 'body' ], "Expected message body" )

        # Super important: makes sure going to right person
        # This is for the metareview, so the receipient should be the REVIEWER
        self.studentRepo.get_student.assert_called_with( self.reviewer.id )

    @patch( 'CanvasHacks.Messaging.SendTools.ConversationMessageSender.send' )
    def test_notify( self, sendMock ):
        sendMock.return_value = 'taco'
        self.obj = DiscussionReviewInvitationMessenger( self.unit, self.studentRepo, self.contentRepo, self.statusRepo )

        # Call
        result = self.obj.notify( [ self.review_assign ], send=True )

        # Check
        # Intermediate things were called
        self.assertTrue( self.contentRepo.get_formatted_work_by.called, "Content repo method called" )
        self.assertTrue( self.studentRepo.get_student.called, "Student repo method called" )
        # Returned expected stuff
        self.assertTrue( len( result ) == 1, "Something was returned" )

        # Check that the sender was given the expected content
        sendMock.assert_called()
        kwargs = sendMock.call_args[ 1 ]
        self.assertEqual( kwargs[ 'student_id' ], self.reviewer.id, "Sent to reviewer" )
        self.assertEqual( kwargs[ 'subject' ], self.activity.email_subject, "Sent with expected subject line" )

        d = self.obj._make_template_input( self.work, None, self.reviewer )
        b = DISCUSSION_REVIEW_NOTICE_TEMPLATE.format( **d )
        self.assertEqual( kwargs[ 'body' ], b, "Sent with expected body" )

        # Super important: makes sure going to right person
        # This is for the metareview, so the receipient should be the REVIEWER
        self.studentRepo.get_student.assert_called_with( self.reviewer.id )



"""
Created by adam on 2/23/20
"""
import pandas as pd

import CanvasHacks.testglobals
from CanvasHacks.Errors.data_ingestion import NoNewSubmissions
from CanvasHacks.Errors.review_pairings import AllAssigned, NoAvailablePartner
# from CanvasHacks.Repositories.factories import WorkRepositoryLoaderFactory
from CanvasHacks.Logging.review_pairings import make_review_audit_file
from CanvasHacks.Logging.run_data import RunLogger
from CanvasHacks.Messaging.skaa import PeerReviewInvitationMessenger
from CanvasHacks.Repositories.status import InvitationStatusRepository
from CanvasHacks.SkaaSteps.ISkaaSteps import IStep

__author__ = 'adam'

from CanvasHacks.Repositories.factories import WorkRepositoryLoaderFactory


class SendInitialWorkToReviewer( IStep ):

    def __init__( self, course=None, unit=None, is_test=None, send=True, **kwargs ):
        """
        :param course:
        :param unit:
        :param is_test:
        :param send: Whether to actually send the messages
        """
        super().__init__( course, unit, is_test, send, **kwargs )

        # The activity whose results we are going to be doing something with
        self.activity = unit.initial_work

        # The activity which we are inviting the receiving student to complete
        self.activity_notifying_about = unit.review

        # The activity whose id is used to store review pairings for the whole SKAA
        self.activity_for_review_pairings = unit.initial_work

        self._initialize()

        # Initialize the relevant status repo
        self.invite_status_repo = InvitationStatusRepository( self.dao, self.activity_notifying_about )

        self.statusRepos = [ self.invite_status_repo]


    def run( self, **kwargs ):
        """
        Loads content assignments, assigns reviewers, and sends formatted
        work to reviewer
        Possible kwargs:
            only_new=False, Probably not used
            rest_timeout=5 Number of seconds to rest_timeout for canvas to generate report
        :return:
        """
        try:
            self._load_step( **kwargs )

            self._assign_step()

            self._message_step()

        except NoNewSubmissions:
            # Check if new submitters, bail if not
            print( "No new submissions" )
            # todo Log run failure
            RunLogger.log_no_submissions( self.unit.initial_work )
            if CanvasHacks.testglobals.TEST:
                # Reraise so can see what happened for tests
                raise NoNewSubmissions

        except AllAssigned as e:
            # Folks already assigned to review
            # have somehow gotten past the new submissions filter
            # This could be due to them resubmitting late
            print( "New submitters but everyone already assigned" )
            # print( e.submitters )
            if CanvasHacks.testglobals.TEST:
                # Reraise so can see what happened for tests
                raise AllAssigned( e.submitters )

        except NoAvailablePartner as e:
            # We will probably want to notify the student that they
            # have had their work noticed, but they are now waiting for
            # a partner
            # todo Notify student that they are waiting
            print( "1 student has submitted and has no partner ") #, e.submitters )
            if CanvasHacks.testglobals.TEST:
                # Reraise so can see what happened for tests
                raise NoAvailablePartner( e.submitters )

    def _message_step( self ):
        # Send the work to the reviewers
        # Note that we still do this even if send is false because
        # the messenger will print out the messages rather than sending them
        self.messenger = PeerReviewInvitationMessenger( self.unit, self.studentRepo, self.work_repo, self.statusRepos )
        # NB, we don't use associationRepo.data because we only
        # want to send to people who are newly assigned
        self.messenger.notify( self.associations, self.send )
        # self.messenger.notify( self.associationRepo.data, self.send )
        #  todo Want some way of tracking if messages fail to send so can resend
        # Log the run
        msg = "Created {} peer review assignments \n {}".format( len( self.associations ), self.associations )
        RunLogger.log_reviews_assigned( self.activity_notifying_about, msg )

    def _assign_step( self ):
        # Assign reviewers to each submitter and store in db
        # NB, the repository will take care of removing already assigned
        # reviewers, so we just give them all submitters
        self.associations = self.associationRepo.assign_reviewers( self.work_repo.submitter_ids )
        if not self.is_test:
            # Save a more readable copy of all the assignments
            # to file
            make_review_audit_file( self.associationRepo, self.studentRepo, self.unit )

    def _load_step( self, **kwargs ):
        self.work_repo = WorkRepositoryLoaderFactory.make( self.activity, self.course, **kwargs )
        self.display_manager.initially_loaded = self.work_repo.data

        # hotfix needed to filter unsubmitted for non quiz ca
        self.work_repo.data = self.work_repo.data[ self.work_repo.data.workflow_state != 'unsubmitted' ]
        if isinstance( self.work_repo.data, pd.DataFrame ):
            self.work_repo.data = self.work_repo.data[ ~pd.isnull( self.work_repo.data.body ) ]

        self.display_manager.post_filter = self.work_repo.data

    @property
    def audit_frame( self ):
        return self.associationRepo.audit_frame( self.studentRepo )


if __name__ == '__main__':
    pass



"""
Created by adam on 2/9/20
"""
from CanvasHacks.Errors.messaging import MessageSendError
from CanvasHacks.Logging.decorators import log_message
from CanvasHacks.Messaging.interfaces import ISender
from CanvasHacks.Api.RequestTools import send_post_request

__author__ = 'adam'

if __name__ == '__main__':
    pass


class ConversationMessageSender( ISender ):
    """Handles sending messages through canvas's conversation
    interface.
    Doing via class allows easier mocking
    """

    def __init__( self ):
        self.url = 'https://canvas.csun.edu/api/v1/conversations'
        self.sent_count = 0
        self.errors = []

    @property
    def error_count( self ):
        return len(self.errors)

    def make_conversation_data( self, student_id, subject, body ):
        """Creates the request data to be sent to canvas"""
        return {
            'recipients': [ student_id ],
            'body': body,
            'subject': subject,
            'force_new': True
        }

    @log_message
    def send( self, student_id, subject, body ):
        """Sends a new message to the student.
       Returns the result object which will contain the conversation id
       if needed for future use
       """
        try:
            d = self.make_conversation_data( student_id, subject, body )
            result = send_post_request( self.url, d )
            self.sent_count += 1
            return result

        except Exception as e:
            self.errors.append(e)
            raise MessageSendError(e)


# ---------------------- old

def make_conversation_data( student_id, subject, body ):
    """Creates the request data to be sent to canvas"""
    return {
        'recipients': [ student_id ],
        'body': body,
        'subject': subject,
        'force_new': True
    }


@log_message
def send_message_to_student( student_id, subject, body ):
    """Sends a new message to the student.
    Returns the result object which will contain the conversation id
    if needed for future use
    """
    d = make_conversation_data( student_id, subject, body )

    return send_post_request( 'https://canvas.csun.edu/api/v1/conversations', d )



"""
Created by adam on 9/20/18
"""
__author__ = 'adam'

from CanvasHacks import environment


# import configparser

# # Load api token and section numbers
# print("Reading credentials from %s" % environment.CREDENTIALS_FILE)
# config = configparser.ConfigParser()
# config.read( environment.CREDENTIALS_FILE)
#
# URL_BASE = config['url'].get('BASE')


def make_url(section_id, verb):
    """Returns the canvas request url"""
    return "%s/%s/%s" % (environment.CONFIG.canvas_url_base, section_id, verb)


if __name__ == '__main__':
    pass


"""
Tests that the correct content will get
sent to the reviewer after the metareview is complete
Created by adam on 2/22/20
"""
__author__ = 'adam'
# todo This should be an end-to-end integration test which actually hits the server

# if __name__ == '__main__':
#     pass
#
# from CanvasHacks.SkaaSteps.SendMetareviewToReviewer import SendMetareviewToReviewer
#
#
#
# # Make sure correct students involved (have more than the 2 in db)
# # Make sure correct text is sent
# # Make sure addressing params are correct
#
#
#
# class TestOnTimeSubmissions(  TestingBase ):
#     """Checks that works properly on first run after
#     deadline on work that has been submitted
#     """
#
#     def setUp(self):
#         self.config_for_test()
#         self.dao = SqliteDAO()
#
#     def test_sends_correctly( self ):
#         """Check that each student receives the expected message
#         containing the correct student's submission
#         """
#         # Load initial content unit data
#
#         # create review assignments
#
#         # send
#
#
#         self.fail()
#
#
#     def test_assignments_stored_correctly( self ):
#         """Check that the students who have submitted the unit
#         (and none of the non-submitteers) have been paired up
#         in the expected manner, and that the pairings have been stored
#         """
#         self.fail()
#
#
# class TestLateSubmissions( unittest.TestCase ):
#     """Checks that works properly on subsequent runs after the
#     initial assignments have been sent out
#
#     """
#
#     def test_sends_correctly( self ):
#         """Check that each student receives the expected message
#         containing the correct student's submission
#         """
#         self.fail()



"""
Created by adam on 5/6/19
"""
from unittest import TestCase

from CanvasHacks.Models.QuizModels import QuizData

__author__ = 'adam'

if __name__ == '__main__':
    pass


class TestQuizData( TestCase ):
    def test_course_id( self ):
        self.skipTest('todo ')

    def test_due_date( self ):
        self.skipTest('todo ')

    def test_quarter_credit_date( self ):
        self.skipTest( 'todo ' )

    def test_quarter_credit_date( self ):
        self.skipTest('todo ')

    def test_lock_date( self ):
        self.skipTest('todo ')

    def test_name( self ):
        self.skipTest('todo ')

    def test_set_question_columns( self ):
        self.skipTest('todo ')

    def test__detect_question_columns( self ):
        test = [ 'submitted', 'attempt', "1785114: \nWhat is an example of persuasive advertising?", '1.0' ]
        qd = QuizData()
        self.assertEqual( qd._detect_question_columns( test ),
                          [ "1785114: \nWhat is an example of persuasive advertising?" ] )



"""
Loggers for simple data about when ran, how many students affected etc
Created by adam on 2/25/20
"""
import CanvasHacks.environment as env
from CanvasHacks.Logging.Logging import TextLogger
from CanvasHacks.TimeTools import current_utc_timestamp

__author__ = 'adam'


class RunLogger( TextLogger ):
    """Logs simple things like when ran etc"""
    log_file_name = env.RUN_LOGNAME

    entry_template = """ 
        \n --------- {time} --------- \n
        Checked for new submissions on activity_inviting_to_complete: {activity_name}
        {message}
        \n -------------------------- \n
        """

    @classmethod
    def log_no_submissions( cls, submitted_activity, time_to_record=None ):
        if time_to_record is None:
            time_to_record = current_utc_timestamp()
        d = {
            'activity_name': submitted_activity.name,
            'time': time_to_record,
            'message': 'No new submissions. No actions taken.'
        }
        cls.write( cls.entry_template.format( **d ) )

    @classmethod
    def log_reviews_assigned( cls, assigned_activity, message, time_to_record=None ):
        if time_to_record is None:
            time_to_record = current_utc_timestamp()
        d = {
            'activity_name': assigned_activity.name,
            'time': time_to_record,
            'message': message
        }
        cls.write( cls.entry_template.format( **d ) )

    @classmethod
    def log_review_feedback_distributed( cls, assigned_activity, message, time_to_record=None ):
        if time_to_record is None:
            time_to_record = current_utc_timestamp()
        d = {
            'activity_name': assigned_activity.name,
            'time': time_to_record,
            'message': message
        }
        cls.write( cls.entry_template.format( **d ) )

    @classmethod
    def log_metareview_feedback_distributed( cls, assigned_activity, message, time_to_record=None ):
        if time_to_record is None:
            time_to_record = current_utc_timestamp()
        d = {
            'activity_name': assigned_activity.name,
            'time': time_to_record,
            'message': message
        }
        cls.write( cls.entry_template.format( **d ) )



if __name__ == '__main__':
    pass



"""
Created by adam on 2/25/20
"""
from datetime import datetime

__author__ = 'adam'

if __name__ == '__main__':
    pass


def entry_separator():
    """Standard separator between log entries for all text logs"""
    return '\n --------------------------------- {} --------------------------- \n'.format( datetime.now() )


def error_entry_separator():
    """Standard separator between log entries for all text logs"""
    return '\n ===================== ERROR ===== {} ===== ERROR ===================== \n'.format( datetime.now() )


import unittest
import datetime

from CanvasHacks.Files.FromDownloadFolder import get_unit_number, is_csv, is_report, get_activity_type,\
    report_file_iterator, DOWNLOAD_FOLDER, file_reports
from CanvasHacks.PeerReviewed.Definitions import TopicalAssignment, Review, MetaReview, DiscussionReview


class TestUtilities( unittest.TestCase ):
    def test_is_csv( self ):
        cases = [ ('Unit 3_ Peer review Survey Student Analysis Report(2).csv', True),
                  ('Unit 3_ Warm up Quiz Student Analysis Report.csv', True),
                  ('Unit 3 stuff in pdf.pdf', False)
                  ]
        for case, expect in cases:
            self.assertEqual( expect, is_csv( case ) )

    def test_is_report( self ):
        cases = [
            ('Unit 3_ Discussion review Survey Student Analysis Report(1).csv', True),
            ('2020-03-15-Unit 3 Metareview-student-work.csv', True),
            ('Unit 34_ Peer review Survey Student Analysis Report(2).csv', True),
            ('Unit 3_ Warm up Quiz Student Analysis Report.csv', True),
            ('Taco truck report 3.csv', False),
            ('TacoUnit 23 report.pdf', False)
        ]
        for case, expect in cases:
            self.assertEqual( expect, is_report( case ), case )

    def test_get_unit_number( self ):
        cases = [
            ('Unit 3_ Discussion review Survey Student Analysis Report(1).csv', 3),
            ('2020-03-15-Unit 13 Metareview-student-work.csv', 13),
            ('Unit 34_ Peer review Survey Student Analysis Report(2).csv', 34),
            ('Unit 1_ Warm up Quiz Student Analysis Report.csv', 1),
            ('UNIT 1_ Warm up Quiz Student Analysis Report.csv', 1),
        ]
        for case, expect in cases:
            self.assertEqual( expect, get_unit_number( case ), case )

    def test_get_activity_type( self ):
        cases = [
            ('Unit 3_ Discussion review Survey Student Analysis Report(1).csv', DiscussionReview),
            ('2020-03-15-Unit 13 Metareview-student-work.csv', MetaReview),
            ('Unit 34_ Peer review Survey Student Analysis Report(2).csv', Review),
            ('Unit 1_ Warm up Quiz Student Analysis Report.csv', TopicalAssignment),
            ('UNIT 1_ Warm up Quiz Student Analysis Report.csv', TopicalAssignment),
        ]
        for case, expect in cases:
            self.assertEqual( expect, get_activity_type( case ), case )

    def test_report_file_iterator( self ):
        fiter = report_file_iterator(DOWNLOAD_FOLDER)
        # NB will not work if no files in actual download folder
        try:
            while True:
                f = next(fiter)
                # check correct keys
                self.assertIn('file_name', f.keys())
                self.assertIn('path', f.keys())
                self.assertIn('activity', f.keys())
        except StopIteration:
            pass

    def test_file_reports( self ):
        self.skipTest("Will actually run the program")
        # reports = file_reports(DOWNLOAD_FOLDER)
        # print(reports)
        # self.assertTrue(len(reports) > 0)


if __name__ == '__main__':
    unittest.main()



"""
Created by adam on 3/2/20
"""
__author__ = 'adam'

from faker import Faker

fake = Faker()
from unittest.mock import MagicMock

from CanvasHacks.Repositories.discussions import DiscussionRepository
from factories.PeerReviewedFactories import unit_factory
from tests.TestingBase import TestingBase

if __name__ == '__main__':
    pass


def discussion_repo_data_factory( students, per_student=1 ):
    out = [ ]
    for _ in range( 0, per_student ):
        for student in students:
            out.append(
                { 'student_id': student.student_id,
                  'student_name': student.name,
                  'text': fake.text()
                  } )
    return out


class TestDiscussionRepository( TestingBase ):
    def setUp( self ):
        self.config_for_test()
        self.unit = unit_factory()
        self.course = MagicMock()
        self.create_new_and_preexisting_students()

        self.activity = self.unit.discussion_forum
        self.obj = DiscussionRepository( activity=self.activity, course=self.course )

    def test_course_id( self ):
        self.skipTest( 'todo' )

    def test_download( self ):
        self.skipTest( 'todo' )

    def test__get_submissions( self ):
        self.skipTest( 'todo' )

    def test__parse_posts_from_submissions( self ):
        self.skipTest( 'todo' )

    def test_get_student_posts( self ):
        self.skipTest( 'todo' )

    def test_get_formatted_work( self ):
        self.skipTest( 'todo' )

    def test_upload_student_grade( self ):
        self.skipTest( 'todo' )

    def test_display_for_grading( self ):
        self.skipTest( 'todo' )

    def test_post_counts( self ):
        self.skipTest( 'todo' )

    def test_filter_by_count( self ):
        min_posts = fake.random.randint(2, 10)
        half_idx = round(len(self.new_students)/2)
        equal_to = discussion_repo_data_factory(self.new_students[half_idx :], min_posts)
        gtr = discussion_repo_data_factory(self.new_students[ : half_idx + 1 ], min_posts)

        to_remain = equal_to + gtr
        to_remove = discussion_repo_data_factory(self.preexisting_students, min_posts -1)

        self.obj.data = to_remove + to_remain

        # call
        result = self.obj.filter_by_count(min_posts)

        # check
        self.assertEqual(len(result), len(to_remain), "expected length returned")

        for s in to_remain:
            self.assertIn(s, result, "expected entries in returned")

        for s in to_remove:
            self.assertNotIn(s, result, "removed entries not in returned")



"""
Created by adam on 12/24/19
"""
__author__ = 'adam'

from faker import Faker

fake = Faker()

import random

import pandas as pd


from CanvasHacks.Models.student import Student


def student_factory():
    name = "{}, {}".format(fake.last_name(), fake.first_name())
    s = Student(student_id=random.randint( 11111, 999999 ), name=name)
    s.sortable_name = name
    return s


def make_students(number):
    """
    Creates a list of student object
    :param number: How many to create
    :return:
    """
    return [student_factory() for _ in range(0, number)]

if __name__ == '__main__':
    pass



"""
Defines things which raise or lower a students' score
on one assignment by how they were rated by another
student on a different assignment

Created by adam on 3/16/20
"""
__author__ = 'adam'

from CanvasHacks.DAOs.mixins import DaoMixin
from CanvasHacks.GradingCorrections.base import IGradeCorrection
from CanvasHacks.GradingMethods.base import IGradingMethod
from CanvasHacks.Models.model import StoreMixin
from CanvasHacks.Definitions.activity import Activity
from CanvasHacks.Repositories.factories import WorkRepositoryLoaderFactory
from CanvasHacks.Repositories.reviewer_associations import AssociationRepository
from CanvasHacks import environment


class ReviewBased( IGradingMethod, StoreMixin, DaoMixin ):

    def __init__( self, graded_activity: Activity, review_activity: Activity, review_columns: list, pct_of_score=1, **kwargs ):
        """

        :param graded_activity: Activity whose grade depends on review
        :param review_activity: Activity where the reviewer gave scores
        :param review_columns: Columns in the review activity to use
        """
        self.pct_of_score = pct_of_score
        self.graded_activity = graded_activity
        self.activity = graded_activity
        self.review_columns = review_columns
        self.review_activity = review_activity


        # todo make this a param or variable...
        self.valmap = {
            'strongly agree' : 1,
            'agree' : 0.9,
            'disagree': 0.5,
            'strongly disagree': 0.1
        }

        self.handle_kwargs(**kwargs)

        self._initialize()


    def _initialize( self ):
        self._initialize_db()
        self.pairingsRepo = AssociationRepository(self.dao, self.graded_activity)
        self.review_repo = WorkRepositoryLoaderFactory.make(course=environment.CONFIG.course,
                                                            activity=self.review_activity,
                                                            rest_timeout=5)


    def grade( self, student_id ):
        # todo assume only one column for now
        col = self.review_columns[0]
        ra = self.pairingsRepo.get_by_author(student_id)
        reviewer_work = self.review_repo.get_student_work(ra.assessor_id)
        v = reviewer_work[col] #.value

        # reformat so don't blow up if i use different caps
        v = v.strip().lower()

        return self.valmap.get(v)


if __name__ == '__main__':
    pass



"""
Created by adam on 5/6/19
"""
from CanvasHacks.Errors.grading import NonStringInContentField
from CanvasHacks.GradingHandlers.base import IGrader
from CanvasHacks.GradingMethods.base import IGradingMethod
from CanvasHacks.GradingCorrections.penalities import get_penalty, IPenalizer
from CanvasHacks.Repositories.interfaces import ISubmissionRepo
from CanvasHacks.Repositories.quizzes import QuizRepository

__author__ = 'adam'

import pandas as pd

if __name__ == '__main__':
    pass


class QuizGrader( IGrader ):

    grade_method: IGradingMethod
    penalizer: IPenalizer

    def __init__( self, work_repo: QuizRepository, submission_repo: ISubmissionRepo, **kwargs ):
        """

        :param work_repo:
        :param submission_repo:
        :param kwargs:
        """
        self.work_repo = work_repo
        self.submission_repo = submission_repo
        super().__init__( **kwargs )

        self.penalizer = self.activity.penalizer
        self.grade_method = self.activity.grade_method

        self.graded = [ ]

    def grade( self, **kwargs):
        """
        Grades all rows in work_repo.data
        todo: Add logging of details of how grade assigned
        :return: List of objects ready for upload
        """

        for i, row in self.work_repo.data.iterrows():
            try:
                self.graded.append( self._grade_row( row, **kwargs ) )
            except NonStringInContentField as e:
                print(e, row)

        # Print or log any penalties applied.
        # The penalizer object leaves this task up to us
        self.report_late_penalties()

        return self.graded

    def report_late_penalties( self ):
        # Report late penalties
        if len( self.penalizer.penalized_records ) > 0:
            for penalty_dict in self.penalizer.penalized_records:
                self._penalty_message(penalty_dict['penalty'], penalty_dict['record'])

    def _get_score( self, content, on_empty=None ):
        """
        Calls the grading method which calculates the points received for
        a given question.
        Abstracted out so that can give a custom response value when
        the grade method returns None
        :param content:
        :return: integer points, None, or on_empty
        """
        g = self.grade_method.grade(content)
        if g:
            # if pd.isnull( row[ column_name ] ):
            return self.work_repo.points_per_question
        elif on_empty is not None:
            return on_empty

    def _make_graded_row_output( self, row, question_score_dict, fudge_points ):
        out = {
            'student_id': int( row[ 'student_id' ] ),
            'attempt': int( row[ 'attempt' ] ),
            'submission_id': int( row[ 'submission_id' ] ),
            'course_id': int( row[ 'course_id' ] ),
            'quiz_id': int( row[ 'quiz_id' ] ),
            'data': { }
        }

        out[ 'data' ][ "quiz_submissions" ] = [
            {
                "attempt": int( row[ 'attempt' ] ),
                "fudge_points": fudge_points,
                "questions": question_score_dict
            }
        ]

        return out

    def _grade_row( self, row , **kwargs):
        """Grades a row

        NB, requires that the non-graded attempts be
        filtered out before passing in

        todo test whether the problem with non-graded people getting zeros is caused here

        :param row: pd.DataFrame row
        :param kwargs:
        :return: Dictionary formatted for uploading
        """
        # used for computing penalty
        total_score = 0
        questions = { }

        # Grade on emptiness
        for qid, column_name in self.work_repo.question_columns:
            content = row[ column_name ]
            pts = self._get_score( content, **kwargs )
            # Quiz type things require the scores to be uploaded
            # for each question.
            # Thus we grade each question and then penalize the
            # overall score with fudge points if necessary
            questions[ qid ] = { 'score': pts }
            total_score += pts

            # Compute penalty if needed
            # Will be 0 if not docking for lateness
            # Records of penalty will be stored on self.penalizer.penalized_records
            fudge_points = self.penalizer.get_fudge_points(row['submitted'], total_score, row)

        out = self._make_graded_row_output(row, questions, fudge_points)

        return out

    def _penalty_message( self, penalty, row ):
        """
        Handles printing or logging of penalties applied

        # todo enable logging

        :param penalty:
        :param row:
        :return:
        """
        stem = 'Student #{}: Submitted on {}; was due {}. Penalized {}'
        return stem.format( row[ 'student_id' ], row[ 'submitted' ], self.activity.due_at, penalty )





# ============================= OLD


def grade( frame, quiz_data_obj, grace_period=None ):
    """
    OLD
    This handles the actual grading

    quiz_data_obj will have the payload format:
        "quiz_submissions": [{
        "attempt": int(attempt),
        "fudge_points": total_score
      },
          "questions": {
      "QUESTION_ID": {
        "score": null, // null for no change, or an unsigned decimal
        "comment": null // null for no change, '' for no comment, or a string
      }
    """
    results = [ ]
    #     questions = detect_question_columns(frame.columns)

    for i, row in frame.iterrows():
        fudge_points = 0
        out = {
            'student_id': int( row[ 'student_id' ] ),
            'attempt': int( row[ 'attempt' ] ),
            'submission_id': int( row[ 'submission_id' ] ),
            'course_id': int( row[ 'course_id' ] ),
            'quiz_id': int( row[ 'quiz_id' ] ),
            'data': { }
        }
        # used for computing penalty
        total_score = 0
        questions = { }

        for qid, column_name in quiz_data_obj.question_columns:
            if pd.isnull( row[ column_name ] ):
                questions[ qid ] = { 'score': 0 }
            else:
                questions[ qid ] = { 'score': 1.0 }
                total_score += 1

        # compute penalty if needed
        penalty = get_penalty( row[ 'submitted' ], quiz_data_obj.due_date, quiz_data_obj.quarter_credit_date,
                               grace_period )
        # will be 0 if not docking
        fudge_points = total_score * -penalty
        if penalty > 0:
            print( 'Student #{}: Submitted on {}; was due {}. Penalized {}'.format( row[ 'student_id' ],
                                                                                    row[ 'submitted' ],
                                                                                    quiz_data_obj.due_date, penalty ) )

        out[ 'data' ][ "quiz_submissions" ] = [
            {
                "attempt": int( row[ 'attempt' ] ),
                "fudge_points": fudge_points,
                "questions": questions
            }
        ]

        results.append( out )
    return results



"""
Created by adam on 12/24/19
"""
from CanvasHacks.Models.model import StoreMixin
from CanvasHacks.Models.student import Student
from CanvasHacks.Definitions.skaa import InitialWork, Review, MetaReview

__author__ = 'adam'

if __name__ == '__main__':
    pass


class Submission( StoreMixin ):
    """A student's completion of an activity_inviting_to_complete
    Parent class; not usually used
    """

    def __init__( self, activity, query_result, **kwargs ):
        self.activity = activity
        # Student who submitted the thing (or thing being reviewed)
        self.submitter = None
        # Student who reviewed the thing
        self.reviewer = None
        # Date the submission was submitted
        self.completion_date = None
        # Score assigned by the reviewer
        self.assigned_score = 0

        self.process( query_result )

        self.handle_kwargs( **kwargs )

    @property
    def on_time( self ):
        """Returns true if the student submitted the activity_inviting_to_complete
        on time.
        """
        return self.completion_date <= self.activity.due_at

    @property
    def submitter_points( self ):
        # todo: Consider adding test to make sure haven't exceeded max score
        # todo: Consider adding check for on time here
        raise NotImplementedError

    def process( self, result ):
        raise NotImplementedError


class InitialSubmission( Submission ):

    def __init__( self, activity, query_result, **kwargs ):
        super().__init__( activity, query_result, **kwargs )

    def process( self, result ):
        """Takes the return from a query getting a peer review
        object and sets values of submission"""
        self.submitter = Student( student_id=result[ 'user_id' ]) #, **result[ 'user' ] )
        self.completion_date = result['submitted_at']
        self.handle_kwargs( **result )

    @property
    def submitter_points( self ):
        # todo checks for minimum length etc
        # todo how to handle late assignments? should do here?
        return self.activity.completion_points


class ReviewSubmission( Submission ):
    def __init__( self, activity, query_result, **kwargs ):
        super().__init__( activity, query_result, **kwargs )

    def process( self, result ):
        """Takes the return from a query getting a peer review
        object and sets values of submission"""
        self.submitter = Student( student_id=result[ 'user_id' ]) #, **result[ 'user' ] )
        self.reviewer = Student( student_id=result[ 'assessor_id' ]) #, **result[ 'assessor' ] )
        self.handle_kwargs( **result )

    @property
    def is_complete( self ):
        return self.workflow_state == 'completed'

    @property
    def reviewer_points( self ):
        """The points the student doing the review should receive"""
        if self.is_complete:
            return self.activity.completion_points
        return 0

    @property
    def submitter_points( self ):
        """The points the student doing the original submission should receive"""
        if self.is_complete:
            # The points that the reviewer assigned
            # todo: make sure can't assign more than assignable points
            return self.assigned_score
        # If the reviewer did not complete the review, the
        # submitting student gets full credit
        # todo: think through the logic here
        return self.activity.assignable_points


class SubmissionFactory( StoreMixin ):
    """A student's completion of an activity_inviting_to_complete
    todo: refactor into factory """

    def __init__( self, activity, **kwargs ):
        self.activity = activity

    def make( self, query_result, **kwargs ):

        if isinstance( self.activity, InitialWork ):
            return InitialSubmission( self.activity, query_result, **kwargs )
        elif isinstance( self.activity, Review ):
            return ReviewSubmission( self.activity, query_result, **kwargs )
        elif isinstance( self.activity, MetaReview ):
            return ReviewSubmission( self.activity, query_result, **kwargs )



"""
Created by adam on 3/22/20
"""
__author__ = 'adam'

from unittest import TestCase

from CanvasHacks.DAOs.sqlite_dao import SqliteDAO
from CanvasHacks.Models.submission_record import SubmissionRecord
from CanvasHacks.Repositories.submission_records import SubmissionRecordRepository
from TestingBase import TestingBase
from factories.PeerReviewedFactories import unit_factory

if __name__ == '__main__':
    pass


class TestSubmissionRecordRepository( TestingBase ):
    def setUp(self):
        self.config_for_test()
        self.unit = unit_factory()
        self.activity = self.unit.discussion_forum
        self.activity_id = self.activity.id

        self.dao = SqliteDAO()
        self.session = self.dao.session
        self.create_new_and_preexisting_students()
        # self.create_preexisting_review_pairings(self.activity_id, self.students)

        # ids of authors previously notified
        self.previously_submitted = []
        # ids of reviewers corresponding to previously notified authors

        self.obj = SubmissionRecordRepository( self.dao, self.activity )


    def test_record( self ):
        # Using id as input
        result = self.obj.record(self.student_ids[0])
        self.assertIsInstance(result, SubmissionRecord, "Correct type returned")
        self.assertEqual(self.student_ids[0], result.student_id, "Correct id set")
        self.assertIsNotNone( result.submitted_at, "Timestamp automatically added" )

        # Using student object as input
        result = self.obj.record(self.students[1])
        self.assertIsInstance(result, SubmissionRecord, "Correct type returned")
        self.assertEqual(self.students[1].id, result.student_id, "Correct id set")
        self.assertIsNotNone( result.submitted_at, "Timestamp automatically added" )



    def test_has_submitted( self ):
        num_previous = 2
        self.make_submitted_records( num_previous )

        others = [sid for sid in self.student_ids if sid not in self.previously_submitted]

        # Check
        for sid in self.previously_submitted:
            self.assertTrue(self.obj.has_submitted(sid), "Returns true for notified students")

        for sid in others:
            self.assertFalse(self.obj.has_submitted(sid), "Returns false for non-notified students")


    def test_previously_submitted( self ):
        num_previous = 2
        self.make_submitted_records( num_previous )

        others = [sid for sid in self.student_ids if sid not in self.previously_submitted]

        # call
        result = self.obj.previously_submitted

        # Check
        for r in result:
            self.assertIsInstance(r, SubmissionRecord, "Returns expected object")
            self.assertIn(r.student_id, self.previously_submitted, "id in previously submitted")
            self.assertNotIn(r.student_id, others, "no unexpected returned")

    def test_previously_submitted_ids( self ):
        num_previous = 2
        self.make_submitted_records( num_previous )
        others = [sid for sid in self.student_ids if sid not in self.previously_submitted]

        # call
        result = self.obj.previously_submitted_ids

        # Check
        for r in result:
            self.assertIsInstance( r, int, "Returns expected type" )
            self.assertIn( r, self.previously_submitted, "id in previously submitted" )
            self.assertNotIn( r, others, "no unexpected returned" )

    def test_get( self ):
        num_previous = 2
        self.make_submitted_records( num_previous )

        # call
        others = [sid for sid in self.student_ids if sid not in self.previously_submitted]

        for sid in others:
            result = self.obj.get(sid)
            self.assertIsNone(result, "Returns none for unsubmitted")

        for sid in self.previously_submitted:
            result = self.obj.get(sid)
            self.assertIsInstance(result, SubmissionRecord, "Correct type returned")
            self.assertEqual(result.student_id, sid, "Correct sid")
            self.assertIsNotNone(result.submitted_at, "Timestamp not empty")

    # def test_reviewers_with_notified_authors( self ):
    #     num_previous = 2
    #     self.make_invitation_submitted_records( num_previous )
    #
    #     reviewers_with_authors_submitted_feedback = [ r.assessor_id for r in self.pairings if
    #                                         r.assessee_id in self.previously_submitted ]
    #
    #     # ra = self.session.query(ReviewAssociation).all()
    #     # fb = self.session.query(FeedbacksubmittedRecord).all()
    #
    #     # try with all students to make it easy to check missing
    #     result = self.obj.reviewers_with_authors_submitted_feedback
    #
    #     self.assertEqual( num_previous, len( result ), "Correct number of records returned" )
    #     for sid in result:
    #         self.assertIn( sid, reviewers_with_authors_submitted_feedback, "Returned for previously notified authors" )
    #
    # def test_remove_reviewers_with_notified_authors( self ):
    #     num_previous = 2
    #     self.make_invitation_submitted_records( num_previous )
    #
    #     reviewers_with_authors_submitted_feedback = [r.assessor_id for r in self.pairings if r.assessee_id in self.previously_submitted]
    #
    #     # ra = self.session.query(ReviewAssociation).all()
    #     # fb = self.session.query(FeedbacksubmittedRecord).all()
    #
    #     # try with all students to make it easy to check missing
    #     result = self.obj.remove_reviewers_with_notified_authors(self.students)
    #
    #     self.assertEqual(len(self.student_ids) - num_previous, len(result), "Correct number of records returned")
    #     for sid in result:
    #         self.assertNotIn(sid, reviewers_with_authors_submitted_feedback, "Author not previously notified")


    # def test_get( self ):
    #     self.fail()
    #
    # def test_has_submitted( self ):
    #     self.fail()
    #
    # def test_submitted_at( self ):
    #     self.fail()
    #
    # def test_previously_submitted( self ):
    #     self.fail()
    #
    # def test_previously_submitted_ids( self ):
    #     self.fail()
    #
    # def test_record( self ):
    #     self.fail()



"""
Created by adam on 2/22/20
"""
__author__ = 'adam'

# todo This should be an end-to-end integration test which actually hits the server

if __name__ == '__main__':
    pass

from CanvasHacks.SkaaSteps.SendReviewToReviewee import SendReviewToReviewee



"""
Created by adam on 3/13/20
"""
__author__ = 'adam'

from CanvasHacks.Messaging.SendTools import ConversationMessageSender

if __name__ == '__main__':
    pass

WAITING_ON_REVIEWER_TEMPLATE = """
Hi {name},

I just wanted to let you know that the person assigned to review your essay assignment has not yet turned it in. 

I will send it to you as soon as they turn it in so that you can do the metareview assignment. 

If they do not turn it in before the metareview assignment is due, you will receive full credit for the metareview.

All best wishes,
/a

"""


class StudentsWaitingOnReviewerMessaging:

    def __init__( self, unit, send=True ):
        self.send = send
        self.unit = unit
        self.activity = self.unit.review
        self.messenger = ConversationMessageSender()
        self.subject = "Waiting for your reviewer to complete the review...".format( self.unit.unit_number )
        self.sent = [ ]


    def prepare_message( self, student_name ):
        data = {
            'name': student_name,
            'assignment_name': self.activity.name,
            # 'original_due_date': self.activity.string_due_date,
            # 'final_date': self.activity.string_lock_date
        }

        return WAITING_ON_REVIEWER_TEMPLATE.format( **data )


    def send_message_to_student( self, student_id, first_name ):
        # student_id = student.id if isinstance(student, User) else student.student_id
        body = self.prepare_message( first_name )
        if self.send:
            msg = self.messenger.send( student_id=student_id, subject=self.subject, body=body )
            self.sent.append( (student_id, msg) )
            print( msg )
        else:
            self.sent.append( (student_id, body) )
            print( body )



"""
Created by adam on 2/10/20
"""
import datetime
import pytz
import pandas as pd

__author__ = 'adam'

if __name__ == '__main__':
    pass


def local_string_to_utc_dt( local_string ):
    """'2020-02-07T07:59:59Z'
    returns Timestamp('2020-02-06 23:59:59-0800', tz='US/Pacific')
    """
    return pd.to_datetime( local_string ).tz_localize('US/Pacific').tz_convert('UTC')


def utc_string_to_local_dt( utc_string ):
    """'2020-02-07T07:59:59Z'
    returns Timestamp('2020-02-06 23:59:59-0800', tz='US/Pacific')
    """
    return pd.to_datetime( utc_string ).tz_convert( 'US/Pacific' )


def check_is_date(date):
    """Checks that a value is a pd.Timestamp
       if not, it tries to make it into one"""
    return date if isinstance( date, pd.Timestamp ) else pd.to_datetime( date )

def current_utc_timestamp():
    return datetime.datetime.now(tz=pytz.utc)

def getDateForMakingFileName():
    return datetime.date.isoformat( datetime.date.today() )

def to_canvas_time(timestamp):
    return timestamp.tz_convert('utc')

def to_csun_time(timestamp):
    return timestamp.tz_convert('US/Pacific')


def timestamp_for_unique_filenames():
    """
    Returns a safe string that can be added to filenames
    to keep them unique, yet still someone human readable
    :return:
    """
    return datetime.datetime.now().strftime( "%Y-%m-%d_%H%M" )


"""
Created by adam on 3/13/20
"""
__author__ = 'adam'
import CanvasHacks.environment as env
from CanvasHacks.DAOs.sqlite_dao import SqliteDAO
import CanvasHacks.testglobals

if __name__ == '__main__':
    pass


class DaoMixin:
    """
    Provides database access initialization.
    Requires that self.unit be set
    """

    def _initialize_db( self ):
        try:
            unit_number = self.unit.unit_number
        except AttributeError:
            unit_number = self.activity.unit_number

        try:
            t = 'TEST-' if env.CONFIG.is_test else ""
            self.db_filepath = "{}/{}{}-Unit-{}-review-assigns.db".format( env.LOG_FOLDER, t, env.CONFIG.semester_name, unit_number )
        except AttributeError as e:
            # This is likely to happen during testing
            print(e)

        if env.CONFIG.is_test:
            try:
                if CanvasHacks.testglobals.TEST_WITH_FILE_DB:
                    # testing: file db
                    self._initialize_file_db()
                    print( "Connected to TEST db file. {}".format( self.db_filepath ) )
                else:
                    # testing: in memory db
                    self._initialize_memory_db()
            except (NameError, AttributeError) as e:
                print(e)
                # The variable might not be defined under in any
                # number of circumstances. So default to the in-memory db
                self._initialize_memory_db()

        else:
            # real: file db
            self._initialize_file_db()
            print( "Connected to REAL db. {}".format( self.db_filepath ) )

    def _initialize_file_db( self ):
        self.dao = SqliteDAO( self.db_filepath )
        self.dao.initialize_db_file()

    def _initialize_memory_db( self ):
        self.dao = SqliteDAO()
        print( "Connected to in-memory testing db" )



"""
Created by adam on 1/21/20
"""

from canvasapi import Canvas
from canvasapi.quiz import QuizReport, Quiz
from canvasapi.requester import Requester
from canvasapi.conversation import Conversation

from CanvasHacks.Models.model import Model

__author__ = 'adam'
import CanvasHacks.environment as env


class StudentUser(Model):

    def __init__(self, token, course_id, **kwargs):
        self.course_id = course_id
        self.token = token
        self.topic_id = None
        self.assignment_id = None
        self.quiz_id = None
        self.discussion_entries = []
        super().__init__(**kwargs)

        self._initialize()

    def _initialize( self ):
        canvas = Canvas(env.CONFIG.canvas_url_base, self.token)
        self.course = canvas.get_course(self.course_id)
        if self.topic_id:
            self.get_discussion(self.topic_id)
        if self.assignment_id:
            self.get_assignment(self.assignment_id)

    def get_discussion( self, topic_id ):
        self.discussion = self.course.get_discussion_topic(topic_id)

    def get_assignment( self, assignment_id ):
        self.assignment = self.course.get_assignment(assignment_id)

    def get_entries( self ):
        """Refresh he list of all entries for he discussion"""
        self.can_entries = self.discussion.get_entries()

    def post_entry( self, message_text ):
        entry = self.discussion.post_entry(message=message_text)
        self.discussion_entries.append(entry)
        # self.discussion_entries[entry.id] = entry
        return entry



if __name__ == '__main__':
    pass


"""
Created by adam on 3/3/20
"""
__author__ = 'adam'


class DisplayManager:
    """Holds data on an iskaa step and handles reporting"""

    def __init__(self, activity):
        self.activity = activity
        # Loading and preparing data
        self._initial_load_count = None
        self._post_filter_count = None

        # Sending notifications
        self._number_to_send = None
        self._number_sent = None

    @property
    def initially_loaded( self ):
        return self._initial_load_count

    @initially_loaded.setter
    def initially_loaded( self, countable ):
        num = countable if isinstance(countable, int) else len(countable)
        self._initial_load_count = num
        print("{} \nLoaded students' work: \t {}".format(self.activity.name, self._initial_load_count ))

    @property
    def number_filtered( self ):
        try:
            return self._initial_load_count - self._post_filter_count
        except Exception as e:
            # print('dm' , e)
            pass

    @property
    def post_filter( self ):
        return self._post_filter_count

    @post_filter.setter
    def post_filter( self, countable ):
        num = countable if isinstance(countable, int) else len(countable)
        self._post_filter_count = num

        print("Previously notified : \t {}".format(self.number_filtered))
        print("Remaining after filtration : \t {}".format(self._post_filter_count))

    @property
    def number_to_send( self ):
        return self._number_to_send

    @number_to_send.setter
    def number_to_send( self, countable):
        num = countable if isinstance( countable, int ) else len( countable )
        self._number_to_send = num

        print("To send: \t {}".format(self._number_to_send))

    @property
    def number_sent( self ):
        return self._number_sent

    @number_sent.setter
    def number_sent( self, countable ):
        num = countable if isinstance( countable, int ) else len( countable )
        self._number_sent = num

        print("Error sending: \t {}".format(self.send_errors_count))
        print("Successfully sent: \t {}".format(self._number_sent))


    @property
    def send_errors_count( self ):
        try:
            return self._number_to_send - self._number_sent
        except Exception as e:
            pass




if __name__ == '__main__':
    pass


"""
Created by adam on 2/27/20
"""
__author__ = 'adam'

if __name__ == '__main__':
    pass


class NonStringInContentField(Exception):
    """Raised when something other than a string
    appears in a field that we are assessing
    """
    pass


class StudentUnableToComplete(Exception):
    """Raised when a student is unable to 
    complete the assignment. Usually because 
    the student they are assigned to review hasn't done their part
    """
    
    def __init__(self, blocked_student, blocking_student):
        self.blocked_student = blocked_student
        self.blocking_student = blocking_student


"""
Tools for selecting whether running on live data or test data
Created by adam on 2/15/20
"""
from IPython.display import display
from ipywidgets import widgets
from CanvasHacks import environment

__author__ = 'adam'


def make_test_selector(return_button=False):
    """
    Displays or returns the button for selecting
    whether we are running on test or live data
    :return:
    """
    def get_description():
        return 'TEST' if environment.CONFIG.is_test else 'LIVE'

    def get_icons():
        return ['check', ''] if environment.CONFIG.is_test else ['', 'check']

    def get_style():
        return 'success' if environment.CONFIG.is_test else 'warning'

    def get_value():
        return environment.CONFIG.is_test

    def get_tooltip():
        return "Click to set to {}".format("TEST" if not environment.CONFIG.is_test else 'LIVE')

    b = widgets.ToggleButton(
        value=get_value(),
        description=get_description(),
        button_style=get_style(),
        tooltip=get_tooltip()
    )
    if return_button is True:
        # If putting inside a box or something else which will
        # call display
        return b
    else:
        display( b )


    def callback(j):
        if not j.new:
            environment.CONFIG.set_live()

        if j.new:
            environment.CONFIG.set_test()

        b.icons = get_icons()
        b.button_style = get_style()
        b.value = get_value()
        b.description = get_description()
        b.tooltip = get_tooltip()

    b.observe(callback, names='value')


if __name__ == '__main__':
    pass


"""
Created by adam on 2/14/19
"""
from CanvasHacks.Definitions.unit import Unit

__author__ = 'adam'
from IPython.display import display
from ipywidgets import widgets, VBox

from CanvasHacks import environment
from CanvasHacks.Api.RequestTools import get_assignments_needing_grading, \
    get_assignments_with_submissions


def make_selection_button( item_id, name, get_func, add_func, remove_func, width='50%', **kwargs ):
    """Creates a single selection button
    style is success if the unit has been selected
    style is primary if not selected
    """

    def get_style( item_id ):
        return 'success' if item_id in get_func() else 'primary'

        # Create the button

    layout = widgets.Layout( width=width )
    b = widgets.Button( description=name, layout=layout, button_style=get_style( item_id ) )

    def callback( change ):
        if item_id in get_func():
            remove_func( item_id )
        else:
            add_func( item_id, name )
        b.button_style = get_style( item_id )

    b.on_click( callback )


    return b


# ------------------------- Discussions
def make_discussion_selection_button( topic_id, name, **kwargs ):
    return make_selection_button( topic_id, name,
                                  environment.CONFIG.get_discussion_ids,
                                  environment.CONFIG.add_discussion,
                                  environment.CONFIG.remove_discussion, **kwargs )


def make_discussion_chooser( course, return_button=False, **kwargs ):
    """Display inputs for selecting assignments
    The selected assignments will be stored in the
    environment.CONFIG
    """
    discussions = [ d for d in course.get_discussion_topics() ]
    # If a particular unit has been selected, just get the component discussions
    # Othewise, just use all course discussions
    if environment.CONFIG.unit_number is not None:
        discussions = environment.CONFIG.unit.find_for_unit(environment.CONFIG.unit_number, discussions)
    buttons = [ ]
    discussions = [ (a.id, a.title) for a in discussions ]
    #     if course_id:
    #         display( widgets.HTML( value="<h4>Course {}</h4>".format( course_id ) ) )
    for discussion_id, discussion_name in discussions:
        buttons.append( make_discussion_selection_button( discussion_id, discussion_name, **kwargs ) )

    if return_button is True:
        # If putting inside a box or something else which will
        # call display
        return buttons
    else:

        display( VBox(buttons) )



# ---------------------------- Assignments

def view_selected_assignments():
    out = widgets.Output( layout={ 'border': '1px solid black' } )
    with out:
        for aid, name in environment.CONFIG.assignments:
            print( name )
    display( out )


def view_ungraded_assignments():
    print( "These assignments need grading: " )
    out = widgets.Output( layout={ 'border': '1px solid black' } )
    to_grade = [ ]
    for sec in environment.CONFIG.course_ids:
        assigns = get_assignments_needing_grading( sec )
        to_grade += [ (g[ 'name' ].strip(), g[ 'id' ]) for g in assigns ]

        # assigns = assigns[ sec ]
        with out:
            to_grade += [ print( g[ 0 ] ) for g in assigns ]
    display( out )


def make_assignment_button( assignment_id, name, **kwargs ):
    """Creates a single selection button
    style is success if the unit has been selected
    style is primary if not selected
    """
    return make_selection_button( assignment_id, name,
                                  environment.CONFIG.get_assignment_ids,
                                  environment.CONFIG.add_assignment,
                                  environment.CONFIG.remove_assignment,
                                  **kwargs )


def make_assignment_chooser(activity=None,  return_button=False, **kwargs):
    """Display inputs for selecting assignments
    The selected assignments will be stored in the
    environment.CONFIG
    """
    assignments = [ ]
    buttons = [ ]
    # Get list of all assignments for the courses
    for course_id in environment.CONFIG.course_ids:
        assignments += get_assignments_with_submissions( course_id )
    assignments = [ (a[ 'id' ], a[ 'name' ]) for a in assignments ]

    if activity is not None:
        # If we we're passed an activity_inviting_to_complete, filter the assignments
        assignments = [a for a in assignments if activity.is_activity_type(a[1])]

    print( "{} assignments with submissions".format( len( assignments ) ) )
    if course_id:
        display( widgets.HTML( value="<h4>Course {}</h4>".format( course_id ) ) )

    # Make buttons for selecting
    for assignment_id, assignment_name in assignments:
        buttons.append( make_assignment_button( assignment_id, assignment_name , **kwargs) )
    # return buttons

    if return_button is True:
        # If putting inside a box or something else which will
        # call display
        return buttons
    else:

        display( VBox(buttons) )


# ------------------------------ Unit

def make_unit_button( unit_number , **kwargs):
    """Creates a single selection button
    style is success if the unit has been selected
    style is primary if not selected
    """
    name = "Unit {}".format( unit_number )

    def set_callback(unit_number, name=None):
        """ This is used so that can also _initialize all the
        canvas objects on the configuration
        """
        environment.CONFIG.set_unit_number(unit_number, name)
        environment.CONFIG.initialize_canvas_objs()
        environment.CONFIG.unit = Unit(environment.CONFIG.course, unit_number)


    return make_selection_button( unit_number,
                                  name,
                                  environment.CONFIG.get_unit_number,
                                  set_callback,
                                  # environment.CONFIG.set_unit_number,
                                  environment.CONFIG.reset_unit_number,
                                  **kwargs
                                  )


def make_unit_chooser( num_units=6, return_button=False, **kwargs ):
    """Display inputs for selecting assignments
    The selected assignments will be stored in the
    environment.CONFIG
    """
    buttons = [ ]
    #     if course_id:
    #         display( widgets.HTML( value="<h4>Course {}</h4>".format( course_id ) ) )
    num_units += 1
    for i in range( 1, num_units ):
        buttons.append( make_unit_button( i, **kwargs ) )

    if return_button is True:
        # If putting inside a box or something else which will
        # call display
        return buttons
    else:

        display( VBox(buttons) )



if __name__ == '__main__':
    pass



"""
Created by adam on 2/23/20
"""
from CanvasHacks.Errors.data_ingestion import NoNewSubmissions
from CanvasHacks.Errors.review_pairings import NoReviewPairingFound
from CanvasHacks.Logging.run_data import RunLogger
from CanvasHacks.Messaging.skaa import MetareviewInvitationMessenger
from CanvasHacks.Repositories.factories import WorkRepositoryLoaderFactory
from CanvasHacks.Repositories.status import FeedbackStatusRepository, InvitationStatusRepository
from CanvasHacks.SkaaSteps.ISkaaSteps import IStep

__author__ = 'adam'


class SendReviewToReviewee( IStep ):
    """Handles loading the submitted reviews and routing them to the authors'
    with instructions for completing the metareview
    """

    def __init__( self, course=None, unit=None, is_test=None, send=True, **kwargs ):
        """
        :param course:
        :param unit:
        :param is_test:
        :param send: Whether to actually send the messages
        """
        super().__init__( course, unit, is_test, send, **kwargs )
        # The activity whose results we are going to be doing something with
        self.activity = unit.review

        # The activity which we are inviting the receiving student to complete
        self.activity_notifying_about = unit.metareview

        # In sending the review results to the author we are
        # telling the about the feedback on the original assignment
        self.activity_feedback_on = unit.initial_work

        # The activity whose id is used to store review pairings for the whole SKAA
        self.activity_for_review_pairings = unit.initial_work

        self.associations = [ ]

        self._initialize()

        # Initialize the relevant status repos

        self.feedback_status_repo = FeedbackStatusRepository(self.dao, self.activity_feedback_on, self.activity_for_review_pairings)

        self.invite_status_repo = InvitationStatusRepository(self.dao, self.activity_notifying_about)

        self.statusRepos = [self.invite_status_repo, self.feedback_status_repo]

    def run( self, **kwargs ):
        """
        Retrieves submitted reviews and sends themto the authors
        along with instructions for metareview
        only_new=False, rest_timeout=5
        :param rest_timeout:
        :param only_new:
        :return:
        """
        try:
            self._load_step( **kwargs )

            self._assign_step()

            self._message_step()

        except NoNewSubmissions:
            # Check if new submitters, bail if not
            print( "No new submissions" )
            # todo Log run failure
            RunLogger.log_no_submissions( self.activity )

        except Exception as e:
            print( e )

    def _message_step( self ):
        # Handle sending the results of the review to the original author
        # so they can do the metareview
        self.messenger = MetareviewInvitationMessenger( self.unit, self.studentRepo, self.work_repo, self.statusRepos )

        self.messenger.notify( self.associations, self.send )

        # Log the run
        msg = "Sent {} peer review results to authors".format( len( self.associations ))
        print(msg)

        lmsg = "{} \n {}".format(msg, self.associations )
        # Note we are distributing the material for the metareview, that's
        # why we're using that activity_inviting_to_complete.
        RunLogger.log_review_feedback_distributed( self.unit.metareview, lmsg )

    def _assign_step( self ):
        self._filter_notified()

        # Filter the review pairs to just those in the work repo.
        for student_id in self.work_repo.submitter_ids:
            try:
                # Work repo contains submitted peer reviews. Thus we look up
                # review pairings where a student submitting the peer review assignment
                # is the assessor
                records = self.associationRepo.get_by_assessor( self.activity_for_review_pairings, student_id )

                if records is None:
                    raise NoReviewPairingFound( student_id )
                else:
                    self.associations.append( records )
            except NoReviewPairingFound:
                pass

        print( "Going to send review results for {} students".format( len( self.associations ) ) )

    def _load_step( self, **kwargs ):
        self.work_repo = WorkRepositoryLoaderFactory.make( self.activity, self.course, **kwargs )
        self.display_manager.initially_loaded = self.work_repo.data

    def _filter_notified( self ):
        """
        Filter out students who have already been invited to
        complete the metareview.
        (NB, a step like this wasn't necessary in SendInitialWorkToReviewer
        since we could filter by who doesn't have a review partner

        :return:
        """
        records = self.feedback_status_repo.reviewers_with_authors_sent_feedback
        self.work_repo.remove_student_records( records )
        self.display_manager.post_filter = self.work_repo.data



if __name__ == '__main__':
    pass



"""
Created by adam on 3/11/20
"""
__author__ = 'adam'

from unittest import TestCase

from CanvasHacks.Processors.cleaners import UtfCleaner
from TestingBase import TestingBase

if __name__ == '__main__':
    pass

TEST_DATA =[ {'input': '¬†¬†¬†¬†¬†¬†¬†¬†¬†¬†¬†',
             'expect': ''
             },
             {'input': 'Good text that ïuh oh.',
              'expect': 'Good text that uh oh.'},
             {'input': "Dog breath is taco-like. Hmmm",
              'expect': "Dog breath is taco-like. Hmmm"
              }
]

class TestUtfCleaner( TestingBase ):

    def setUp(self) -> None:
        self.config_for_test()
        self.obj = UtfCleaner()

    def test_clean( self ):
        for t in TEST_DATA:
            self.assertEqual(t['expect'], self.obj.clean(t['input']))



"""
Tools for reading and writing csv files

Created by adam on 6/22/18
"""
__author__ = 'adam'

import csv
import os


def create_folder( folder_path ):
    try:
        os.mkdir( folder_path )
        print( "created: %s" % folder_path )
    except Exception as e:
        print( "error creating: %s" % folder_path )
        print( e )
        pass


def get_newest_file( folder_path ):
    """Returns the file in the folder with the newest
    creation date
    """
    fiter = makeDataFileIterator( folder_path )
    fpaths = [ (f, os.path.getctime( f )) for f in fiter ]
    fpaths.sort( key=lambda x: x[ 1 ], reverse=True )
    newest = fpaths[ 0 ][ 0 ]
    return newest


def makeDataFileIterator( folderPath, exclude=[ ] ):
    """
    Returns an iterator of all files in the source directory
    so that each file has its path appended to it.

    Args:
        folderPath: The path to get file names from
        exclude: file names which should not be included in the output list

    """
    exclude = exclude if any( exclude ) else [ '.DS_Store' ]
    for root, dirs, files in os.walk( folderPath ):
        for name in files:
            if name not in exclude:
                yield os.path.join( root, name )


def write_csv( csvFile, rowToWrite ):
    """TODO: let this figure out whether needs i, j or i"""
    with open( csvFile, 'a' ) as csvfile:
        writer = csv.writer( csvfile )
        # for i in toWrite:
        if type( rowToWrite ) is not list:
            rowToWrite = [ rowToWrite ]
        writer.writerow( rowToWrite )


#             for i, j in toWrite:
#                 writer.writerow([i, j])

def write_dict_to_csv( csvFile, toWrite ):
    """"Writes a mapping dictionary to a csv file """
    with open( csvFile, 'w' ) as csvfile:
        writer = csv.writer( csvfile )
        for k in toWrite.keys():
            for v in toWrite[ k ]:
                writer.writerow( [ k, v ] )


def read_csv_mapping( csvFile, firstColIsKey=True ):
    """Returns a dictionary mapping values in the first column to values in the second. That way
     the second column list of tuples for each row of the stored item

    Args:
        firstColIsKey (oboolean): Whether the first column in the file is to be used as keys
    """
    with open( csvFile ) as csvFile:
        reader = csv.reader( csvFile, quotechar='|' )
        out = { }
        for row in reader:
            if firstColIsKey:
                out[ row[ 0 ] ] = row[ 1 ]
            else:
                out[ row[ 1 ] ] = row[ 0 ]
        return out


def read_csv( csvFile ):
    """Returns a list of tuples for each row of the stored item"""
    with open( csvFile ) as csvFile:
        reader = csv.reader( csvFile, quotechar='|' )
        out = [ ]
        for row in reader:
            out.append( tuple( row ) )
        return out


def read_list( csvFile ):
    """Returns a list which has been stored as a csv file"""
    with open( csvFile ) as csvFile:
        reader = csv.reader( csvFile, quotechar='|' )
        out = [ ]
        for row in reader:
            out += row
        return out


def read_list_generator( csvFile ):
    """Returns a generator for items in a list which has been stored in a csv file"""
    results = read_list( csvFile )
    for r in results:
        yield r


# Get all subfolder paths

def read_text_block( filename ):
    with open( filename, 'r' ) as f:
        j = f.read()
        j = "\n".join( j.split( '\\n' ) )
        # j = "\".join(j.split("\\")
        return j


if __name__ == '__main__':
    pass



"""
Tools for handling and maintaining who is assigned to review whom

Created by adam on 12/28/19
"""
from CanvasHacks.DAOs.sqlite_dao import SqliteDAO
from CanvasHacks.Errors.review_pairings import NoAvailablePartner, AllAssigned
from CanvasHacks.Models.review_association import ReviewAssociation
from CanvasHacks.Definitions.activity import Activity
import pandas as pd

__author__ = 'adam'
import numpy as np
import random

from CanvasHacks.Repositories.mixins import ObjectHandlerMixin


def assign_reviewers( student_ids ):
    """
    Takes a list of students or student ids and creates a dict
    of reviewers by shifting them all one to the right

    >>>test_ids = [1, 2, 3, 4]
    >>>test_ids = [1, 2, 3, 4, 5, 6, 7]
    >>>r = assign_reviewers(test_ids)
    >>>for a, b in r:
    >>>   assert(a != b) # no id assigned to self
    """
    return list( zip( student_ids, np.roll( student_ids, 1 ) ) )


# def force_to_ids(list_of_students):
#     """We could receive a list of ids, Student
#     objects or canvasapi User objects. This returns
#     a list of ids"""
#     out = []
#
#     for s in list_of_students:
#         if isinstance(s, int):
#             out.append(s)
#         else:
#             # Both Student and User objects will have an id attribute
#             # which contains the canvas id of the student
#             out.append(s.id)
#     return out


class AssociationRepository(ObjectHandlerMixin):

    def __init__( self, dao: SqliteDAO, activity: Activity):
        """
        Create a repository to handle review assignments for a
        particular activity
        """
        self.activity = activity
        self.session = dao.session

    def _make_associations( self, submitters ):
        """Handles creating the review unit associations.
        We use this since it builds in checks for problem cases given that
        for some assignments students may submit more than
        one item."""
        while True:
            # randomize list of submitters
            # to hopefully (further) split up adjacent submissions by the
            # same student
            random.shuffle( submitters )
            # Making a list upon input incase we've been passed an iterator
            candidate = assign_reviewers( [ s for s in submitters ] )
            bad = False
            for b, c in candidate:
                # Perform checks
                # Note that this should work regardless of whether
                # received list of student objects or just their ids
                if b == c:
                    # self unit so start over
                    bad = True
            if bad:
                continue
            else:
                # if we completed the loops, no one is reviewing themselves
                return candidate

    def assign_reviewers( self, submitters ):
        """
        Assigns every student in submitters to review one other
        student.
        This is the main method to call
        :param submitters: List of ids or students
        :return list of ReviewAssociation objects
        :raises AllAssigned, NoAvailablePartner,
        """
        # Force list of submitters to be a list of ids
        submitters = self.force_to_ids(submitters)

        # Filter out anyone who has already been assigned to review
        # someone. That will leave anyone who is just now submitting
        # and anyone who was left hanging without a reviewee in a previous
        # run.
        filtered_submitters = self.filter_assigned_reviewers(submitters)

        # Check whether everyone is already paired up
        if len(filtered_submitters) == 0:
            # We will return the original list because this likely
            # indicates people have done something wierd like resubmit late
            raise AllAssigned(submitters)

        # Check whether we only have 1 new submitter
        # and thus will not be able to partner them up
        if len(filtered_submitters) == 1:
            raise NoAvailablePartner(filtered_submitters)

        # Pair up the remaining students
        assoc_to_make = self._make_associations(  filtered_submitters)

        self.new_assignments = []
        for assessor, assessee in assoc_to_make:
            # print(assessor, assessee)
            a = self._create_association( self.activity, assessor_id=assessor, assessee_id=assessee )
            self.new_assignments.append(a)
        print( '{} Review associations created for {} submitters and stored in db'.format(len(self.new_assignments), len( filtered_submitters ) ) )
        print("See this object's audit_frame method for details")

        # make_review_audit_file(self, self.student_repository)
        return self.new_assignments

    def audit_frame( self, student_repository ):
        """
        Returns a data frame with all the current review
        pairings
        :return: DataFrame
        """
        review_audit = []
        for rev in self.get_associations():
            # print(rev.assessor_id, rev.assessee_id)
            assessor = student_repository.get_student( rev.assessor_id )
            assessee = student_repository.get_student( rev.assessee_id )
            # print(assessor)

            review_audit.append({
                'activity_inviting_to_complete' : self.activity.name,
                'assessor_name' : assessor.short_name,
                'assessor_sis_id': assessor.sis_user_id,
                'assessor_canvas_id': assessor.id,
                'assessee_name' : assessee.short_name,
                'assessee_sis_id': assessee.sis_user_id,
                'assessee_canvas_id': assessee.id,
            })

        return pd.DataFrame(review_audit)

    def _create_association( self, activity, assessor_id, assessee_id ):
        """Creates a ReviewAssociation object for the given students and
        saves it to the db
        Leaving activity_inviting_to_complete as a param even though it is now a property of the object
        so test methods can call on its own
        """
        ra = ReviewAssociation( activity_id=activity.id, assessor_id=int(assessor_id), assessee_id=int(assessee_id) )
        self.session.add( ra )
        self.session.commit()
        return ra

    def get_associations( self, activity=None ):
        """Returns all review assignments for the activity_inviting_to_complete
        Leaving activity_inviting_to_complete as a param even though it is now a property of the object
        so test methods can call on its own """
        if activity is None:
            activity = self.activity
        return self.session.query( ReviewAssociation ) \
            .filter( ReviewAssociation.activity_id == activity.id ) \
            .all()

    def get_assessor_object( self, activity, author_id ):
        """Returns the ReviewAssociation object containing the
        of the student assigned to review the submitter
        or None if no student has been assigned
        Leaving activity_inviting_to_complete as a param even though it is now a property of the object
        so test methods can call on its own
        """
        return self.session.query( ReviewAssociation ) \
            .filter( ReviewAssociation.activity_id == activity.id ) \
            .filter( ReviewAssociation.assessee_id == author_id ) \
            .one_or_none()

    @property
    def data( self):
        return self.get_associations()

    def get_assessor( self, activity, submitter_id ):
        """Returns the id of the student assigned to review the submitter
        or None if no student has been assigned
        Leaving activity_inviting_to_complete as a param even though it is now a property of the object
        so test methods can call on its own
        """
        r = self.get_assessor_object(activity, submitter_id)
        # r = self.session.query( ReviewAssociation ) \
        #     .filter( ReviewAssociation.activity_id == activity_inviting_to_complete.id ) \
        #     .filter( ReviewAssociation.assessee_id == submitter_id ) \
        #     .one_or_none()
        if r:
            return r.assessor_id
        return r

    def get_by_assessor( self, activity, reviewer_id ):
        """
        Returns the review object where the student is the assessor
        :param activity:
        :param reviewer_id:
        :return:
        """
        return self.session.query( ReviewAssociation ) \
            .filter( ReviewAssociation.activity_id == activity.id ) \
            .filter( ReviewAssociation.assessor_id == reviewer_id ) \
            .one_or_none()

    def get_by_assessee( self, activity, author_id ):
        """
        Returns the review object where the student is the author of
        the original work
        :param activity:
        :param author_id:
        :return:
        """
        return self.session.query( ReviewAssociation ) \
            .filter( ReviewAssociation.activity_id == activity.id ) \
            .filter( ReviewAssociation.assessee_id == author_id ) \
            .one_or_none()

    def get_by_author( self, author ):
        """
        Returns the review object where the student is the author of
        the original work
        Introduced in preparation for CAN-52
        :param author: Student or student id
        :return: ReviewAssociation or None
        """
        author_id = self._handle_id(author)
        return self.session.query( ReviewAssociation )\
            .filter( ReviewAssociation.activity_id == self.activity.id )\
            .filter( ReviewAssociation.assessee_id == author_id )\
            .one_or_none()

    def get_by_reviewer( self, reviewer ):
        """
        Returns the review object where the student is the assessor of the
        original work

        Introduced in preparation for CAN-52

        :param reviewer: Student or student id
        :return:
        """
        reviewer_id = self._handle_id(reviewer)
        return self.session.query( ReviewAssociation )\
            .filter( ReviewAssociation.activity_id == self.activity.id )\
            .filter( ReviewAssociation.assessor_id == reviewer_id )\
            .one_or_none()

    def get_assessee_object( self, activity, reviewer_id ):
        """Returns the ReviewAssociation object containing the student
         assigned to be reviewed by the reviewer
        or None if no student has been assigned
        Leaving activity_inviting_to_complete as a param even though it is now a property of the object
        so test methods can call on its own
        """
        return self.session.query( ReviewAssociation ) \
            .filter( ReviewAssociation.activity_id == activity.id ) \
            .filter( ReviewAssociation.assessor_id == reviewer_id ) \
            .one_or_none()

    def get_assessee( self, activity, reviewer_id ):
        """Returns the student assigned to be reviewed by the reviewer
        or None if no student has been assigned
        Leaving activity_inviting_to_complete as a param even though it is now a property of the object
        so test methods can call on its own
        """
        r = self.get_assessee_object(activity, reviewer_id)
        if r:
            return r.assessee_id
        return r

    def get_for_assignment( self, assignment ):
        """Returns list of all submitter, reviewer tuples"""
        pass

    def filter_assigned_reviewers( self, submitters ):
        """Removes entries from list who have already
        been assigned to review someone"""
        prev = len(submitters)
        submitters = [ s for s in submitters if self.get_assessee(self.activity, s) is None]
        print("Removed {} submitters who were already assigned".format(prev - len(submitters) ))
        return submitters


if __name__ == '__main__':
    pass



# """
# Created by adam on 3/8/20
# """
# __author__ = 'adam'
#
# from unittest import TestCase
#
# from CanvasHacks.DAOs.sqlite_dao import SqliteDAO
# from CanvasHacks.Repositories.status import MetareviewStatusRepository, InvitationStatusRepository,\
#     FeedbackStatusRepository
# from TestingBase import TestingBase
# from factories.PeerReviewedFactories import unit_factory
#
# if __name__ == '__main__':
#     pass
#
#
# class TestMetareviewStatusRepository( TestingBase ):
#     def setUp(self):
#         self.config_for_test()
#         self.unit = unit_factory()
#         self.activity = self.unit.discussion_forum
#         self.activity_id = self.activity.id
#
#         self.dao = SqliteDAO()
#         self.session = self.dao.session
#         self.create_new_and_preexisting_students()
#
#         self.obj = MetareviewStatusRepository( self.dao, self.unit )
#
#     def test_initialization( self ):
#         self.assertEqual(self.obj.feedback_activity, self.unit.initial_work, "Recording that feedback on initial work being sent")
#         self.assertIsInstance(self.obj.feedback_repo, FeedbackStatusRepository, "Correct feedback repo instantiated")
#
#         self.assertEqual(self.obj.invitation_activity, self.unit.metareview, "Recording that being invited to complete metareview")
#         self.assertIsInstance(self.obj.invite_repo, InvitationStatusRepository, "Correct feedback repo instantiated")
#
#
#     def test_record( self ):
#         self.fail()



# """
# Created by adam on 3/15/20
# """
# __author__ = 'adam'
#
# from unittest import TestCase
#
# from CanvasHacks.GradingMethods.base import IGradingMethod
# from TestingBase import TestingBase
#
# if __name__ == '__main__':
#     pass
#
#
# class TestIGradingMethod( TestingBase ):
#     def setUp(self) -> None:
#         self.config_for_test()
#
#     def test_to_remove_no_stopwords( self ):
#         obj = IGradingMethod()
#         obj.count_stopwords = True
#         self.assertIn('.', obj.to_remove)
#         self.assertNotIn('the', obj.to_remove)
#
#     def test_to_remove_inc_stopwords( self ):
#         obj = IGradingMethod()
#         obj.count_stopwords = False
#         self.assertIn( '.', obj.to_remove )
#         self.assertIn( 'the', obj.to_remove )
#
#     def test_to_remove_count_stopwords_not_set( self ):
#         obj = IGradingMethod()
#         self.assertEqual([], obj.to_remove, "returns an empty list if attribute missing")
#
#
#     def test__determine_word_count( self ):
#         obj = IGradingMethod()
#         obj.count_stopwords = True
#         expected = self.fake.random.randint(1, 1000)
#         content = " ".join([ self.fake.word() for _ in range(0, expected)])
#
#         result = obj._determine_word_count(content)
#
#         self.assertEqual(expected, result, 'expected count returned')
#



"""
Created by adam on 3/22/20
"""
__author__ = 'adam'

from CanvasHacks.DAOs.sqlite_dao import SqliteDAO
from CanvasHacks.Models.submission_record import SubmissionRecord
from CanvasHacks.Definitions.activity import Activity
from CanvasHacks.Repositories.interfaces import IRepo
from CanvasHacks.Repositories.mixins import StudentWorkMixin
from CanvasHacks.TimeTools import current_utc_timestamp

if __name__ == '__main__':
    pass


class SubmissionRecordRepository( StudentWorkMixin, IRepo ):
    
    def __init__( self, dao: SqliteDAO, activity: Activity ):
        """
        Create a repository to handle events for a
        particular activity_inviting_to_complete
        """
        self.activity = activity
        self.session = dao.session

    def get( self, student ):
        """
        Returns the record for the student for this activity or none
        if no record has been created yet
        :param student: Student or student id
        :return: SubmissionRecord or None
        """
        student_id = self._handle_id( student )
        return self.session.query( SubmissionRecord )\
            .filter( SubmissionRecord.activity_id == self.activity.id )\
            .filter( SubmissionRecord.student_id == student_id )\
            .one_or_none()

    def has_submitted( self, student ):
        """
        Returns true if the student has submitted the activity 
        :param student: Student or student id
        :return:
        """
        record = self.get( student )
        return record is not None

    def submitted_at( self, student ):
        """Returns the timestamp of when the
        student submitted the activity or None if they haven't
        """
        rec = self.get( student )
        if rec is not None:
            return rec.sent_at
        return None

    @property
    def previously_submitted( self ):
        """
        Returns SubmissionRecord for all students
        who have been invited to complete the activity
        :return: list
        """
        return self.session\
            .query( SubmissionRecord )\
            .filter( SubmissionRecord.activity_id == self.activity.id )\
            .all()

    @property
    def previously_submitted_ids( self ):
        """Returns a list of ids of students who have
        already submitted
        """
        recs = self.previously_submitted
        if len( recs ) == 0:
            return recs
        return [ r.student_id for r in recs ]

    def record( self, student, time_to_record=None ):
        """
        Records that a student has submitted  an activity

        # todo Consider bulk commits if slow

        :param student: Student or student id
        :param time_to_record: If none, we will record the current utc time
        :return: SubmissionRecord
        """
        student_id = self._handle_id( student )
        if time_to_record is None:
            time_to_record = current_utc_timestamp()

        rec = SubmissionRecord( student_id=student_id,
                                activity_id=self.activity.id,
                                submitted_at=time_to_record )

        self.session.add( rec )
        self.session.commit()
        return rec



"""
Created by adam on 3/11/20
"""
__author__ = 'adam'

from CanvasHacks.Processors.cleaners import TextCleaner
from TestingBase import TestingBase

if __name__ == '__main__':
    pass

TEST_DATA = [ { 'input': '<div><span="dog">¬†¬†¬†¬†¬†¬†¬†¬†¬†¬†¬†</span></div>',
                'expect': ''
                },
              { 'input': 'Agree',
                'expect': 'Agree' },
              { 'input': 45,
                'expect': 45
                },
              { 'input': '<div><span="dog">Dog breath is taco-like.¬†¬† Hmmm¬†¬†¬†¬†¬†¬†¬†¬†¬†</span></div>',
                'expect': 'Dog breath is taco-like. Hmmm'
                },
              { 'input': "Dog breath is taco-like. Hmmm",
                'expect': "Dog breath is taco-like. Hmmm"
                }
              ]


class TestTextCleaner( TestingBase ):

    def setUp( self ) -> None:
        self.config_for_test()
        self.obj = TextCleaner()

    def test_clean( self ):
        for t in TEST_DATA:
            self.assertEqual( t[ 'expect' ], self.obj.clean( t[ 'input' ] ) )



"""
Created by adam on 3/20/20
"""
__author__ = 'adam'

from unittest import TestCase
from unittest.mock import patch, MagicMock
import pandas as pd

from CanvasHacks.GradingMethods.review import ReviewBased
from TestingBase import TestingBase
from factories.PeerReviewedFactories import unit_factory
from factories.RepositoryMocks import ContentRepositoryMock

if __name__ == '__main__':
    pass


class TestReviewBased( TestingBase ):

    def setUp(self) -> None:
        self.config_for_test()
        self.create_new_and_preexisting_students()
        self.unit = unit_factory()
        self.graded_activity = self.unit.initial_work
        self.review_activity = self.unit.review
        self.column_names = ['taco']

        # self.reviewRepo = ContentRepositoryMock()

    def test__initialize( self ):
        self.skipTest('todo')

    @patch('CanvasHacks.GradingMethods.review.WorkRepositoryLoaderFactory')
    def test_grade( self, workRepoMock ):
        # 'strongly agree': 1,
        # 'agree': 0.9,
        # 'disagree': 0.5,
        # 'strongly disagree': 0.1
        p = {'get_student_work.return_value': pd.DataFrame([
            {'student_id': self.students[0].student_id,
             self.column_names[0]: 'Strongly Agree'
             }]
        ).iloc[0]}

        reviewRepo = MagicMock(**p)
        workRepoMock.make = MagicMock(return_value=reviewRepo)

        self.obj = ReviewBased(graded_activity=self.graded_activity,
                               review_activity=self.review_activity,
                               pct_of_score=1,
                               review_columns=self.column_names)
        self.session = self.obj.dao.session
        self.create_preexisting_review_pairings(activity_id=self.graded_activity.id,
                                                preexisting_students=self.students)

        # call
        result = self.obj.grade(self.students[0].student_id)

        # check
        self.assertEqual(result, 1, "returns expected pct")


    def test_grade_portion_of_total_score( self ):
        self.skipTest('todo')

    def test_grade_reviewing_student_not_turned_in( self ):
        self.skipTest('todo')



"""
Created by adam on 2/24/20
"""
# import CanvasHacks.globals
# CanvasHacks.globals.use_api = False
from unittest.mock import MagicMock

from CanvasHacks.Loaders.assignment import AssignmentDownloadLoader
from CanvasHacks.Loaders.discussion import DiscussionDownloadLoader, DiscussionFileLoader
from CanvasHacks.Loaders.factories import LoaderFactory, QuizLoaderFactory
from CanvasHacks.Loaders.quiz import AllQuizReportDownloader, AllQuizReportFileLoader, NewQuizReportDownloadLoader, \
    NewQuizReportFileLoader
from factories.PeerReviewedFactories import unit_factory
from tests.TestingBase import TestingBase

__author__ = 'adam'


class TestLoaderFactory( TestingBase ):

    def setUp( self ):
        self.config_for_test()
        self.course = MagicMock()
        self.unit = unit_factory()

    def test_returns_discussion_download_loader( self ):
        r = LoaderFactory.make(is_discussion=True)
        self.assertEqual(r, DiscussionDownloadLoader)

    def test_returns_discussion_file_loader( self ):
        r = LoaderFactory.make(is_discussion=True, download=False)
        self.assertEqual(r, DiscussionFileLoader)

    def test_returns_assignment_loader( self ):
        r = LoaderFactory.make( is_quiz=False )
        self.assertEqual( r, AssignmentDownloadLoader )

    # Tests to make sure quiz loader passes correct loader up
    # to the main loader factory

    def test_make_download_new( self ):
        download = True
        only_new = True

        # call
        result = LoaderFactory.make( download, only_new )

        # check
        self.assertEqual( result, NewQuizReportDownloadLoader )

    def test_make_download_all( self ):
        download = True
        only_new = False

        # call
        result = LoaderFactory.make( download, only_new )

        # check
        self.assertEqual( result, AllQuizReportDownloader )

    def test_make_file_new( self ):
        download = False
        only_new = True

        # call
        result = LoaderFactory.make( download, only_new )

        # check
        self.assertEqual( result, NewQuizReportFileLoader )

    def test_make_file_all( self ):
        download = False
        only_new = False

        # call
        result = LoaderFactory.make( download, only_new )

        # check
        self.assertEqual( result, AllQuizReportFileLoader )


class TestAssignmentLoaderFactory( TestingBase ):

    def setUp( self ):
        self.config_for_test()

    def test_returns_assignment_loader( self ):
        r = LoaderFactory.make( is_quiz=False )
        self.assertEqual(r, AssignmentDownloadLoader, "Returns the AssignmentDownloadLoader")


class TestQuizLoaderFactory( TestingBase ):

    def setUp( self ):
        self.config_for_test()

    def test_make_download_new( self ):
        download = True
        only_new = True

        # call
        result = QuizLoaderFactory.make( download, only_new )

        # check
        self.assertEqual( result, NewQuizReportDownloadLoader )

    def test_make_download_all( self ):
        download = True
        only_new = False

        # call
        result = LoaderFactory.make( download, only_new )

        # check
        self.assertEqual( result, AllQuizReportDownloader )

    def test_make_file_new( self ):
        download = False
        only_new = True

        # call
        result = LoaderFactory.make( download, only_new )

        # check
        self.assertEqual( result, NewQuizReportFileLoader )

    def test_make_file_all( self ):
        download = False
        only_new = False

        # call
        result = LoaderFactory.make( download, only_new )

        # check
        self.assertEqual( result, AllQuizReportFileLoader )



"""
Created by adam on 3/9/20
"""
__author__ = 'adam'

from unittest.mock import MagicMock, create_autospec, patch

from CanvasHacks.Definitions.base import BlockableActivity
from CanvasHacks.Errors.grading import StudentUnableToComplete
from CanvasHacks.GradingCorrections.base import IGradeCorrection
from CanvasHacks.GradingHandlers.quiz_new import QuizGrader
from CanvasHacks.GradingMethods.base import IGradingMethod
from CanvasHacks.Repositories.interfaces import ISubmissionRepo
from TestingBase import TestingBase
from factories.RepositoryMocks import ContentRepositoryMock

if __name__ == '__main__':
    pass


class TestQuizGraderUnitTests( TestingBase ):

    def setUp( self ):
        self.num = 3
        self.config_for_test()
        self.points_per_question = self.fake.random.randint( 1, 100 )

        self.create_new_and_preexisting_students()
        sp = { 'analyze.return_value': 0.5 }
        self.penalizers = [ ]
        for _ in range( 0, self.num ):
            m = MagicMock( **sp )
            # m.analyze = MagicMock( return_value=0.5 )
            self.penalizers.append( m )
        self.grade_methods = [ create_autospec( IGradingMethod ) for _ in range( 0, self.num ) ]
        self.corrections = [ create_autospec( IGradeCorrection ) for _ in range( 0, self.num ) ]
        self.activity = MagicMock( grade_methods=self.grade_methods, penalizers=self.penalizers,
                                   corrections=self.corrections )

        self.work_repo = ContentRepositoryMock( activity=self.activity, points_per_question=self.points_per_question )
        self.work_repo.create_quiz_repo_data( self.student_ids, submitted_at=self.fake.date_time_this_century(),
                                              make_dataframe=True )

        self.submission_repo = create_autospec( ISubmissionRepo )

        self.obj = QuizGrader( self.work_repo, self.submission_repo )

    def test__compute_initial_total( self ):
        for i, row in self.work_repo.data.iterrows():
            # call
            results = self.obj._compute_initial_total( row )
            # check
            for gm in self.grade_methods:
                gm.grade.assert_called()
                # gm.grade.assert_called_with(content)
                for qid, column_name in self.work_repo.question_columns:
                    gm.grade.assert_any_call( row[ column_name ] )

            # print(results)

    def test__compute_correction_pct( self ):
        for i, row in self.work_repo.data.iterrows():
            # call
            results = self.obj._compute_correction_pct( row )
            # check
            for gm in self.corrections:
                gm.analyze.assert_called()
                # gm.grade.assert_called_with(content)
                # for qid, column_name in self.work_repo.question_columns:
                # check called with row
                gm.analyze.assert_any_call( **row.to_dict() )

    def test__compute_penalty_pct( self ):
        """
        Since the penalizer will mostly handle penalties based
        on times, we check that particular use case.
        :return:
        """
        for i, row in self.work_repo.data.iterrows():
            # call
            results = self.obj._compute_penalty_pct( row )
            # check
            for gm in self.penalizers:
                gm.analyze.assert_called()
                gm.analyze.assert_called_with( **row.to_dict() )

    def test__compute_penalty_pct_specific_to_times( self ):
        """
        Since the penalizer will mostly handle penalties based
        on times, we check that particular use case.
        :return:
        """
        for i, row in self.work_repo.data.iterrows():
            # call
            results = self.obj._compute_penalty_pct( row )
            # check
            for gm in self.penalizers:
                gm.analyze.assert_called()
                call_args = gm.analyze.call_args
                s = call_args[ 1 ][ 'submitted_at' ]
                self.assertEqual( row[ 'submitted_at' ], s )

    # def test__get_score_custom_on_empty( self ):
    #     # Custom on empty case
    #     self.grade_method.grade = MagicMock( return_value=None )
    #     content = self.fake.paragraph()
    #     on_empty = 'taco'
    #
    #     # call
    #     result = self.obj._get_score( content, on_empty=on_empty )
    #
    #     # check
    #     self.assertEqual(on_empty, result, "returns custom value when empty")
    #     self.grade_method.grade.assert_called()
    #     self.grade_method.grade.assert_called_with( content )


class TestQuizGraderFunctionalTests( TestingBase ):

    def setUp( self ):
        # each grade method needs to return .25
        self.num = 4
        self.config_for_test()
        self.points_per_question = self.fake.random.randint( 1, 100 )

        self.create_new_and_preexisting_students()
        # sp = { 'analyze.return_value': 0.5 }
        self.penalizers = [ ]
        # for _ in range( 0, self.num ):
        #     m = MagicMock( **sp )
        #     self.penalizers.append( m )
        self.grade_methods = []
        gp = { 'grade.return_value': 0.25 }
        for _ in range( 0, self.num ):
            self.grade_methods.append( MagicMock( **gp ) )

        # self.grade_methods = [ create_autospec( IGradingMethod ) for _ in range( 0, self.num ) ]
        sp = { 'analyze.return_value': 0 }
        self.corrections = [ MagicMock(**sp) for _ in range( 0, self.num ) ]

        self.activity = MagicMock( grade_methods=self.grade_methods,
                                   penalizers=self.penalizers,
                                   corrections=self.corrections )

        self.work_repo = ContentRepositoryMock( activity=self.activity, points_per_question=self.points_per_question )
        self.work_repo.create_quiz_repo_data( self.student_ids, submitted_at=self.fake.date_time_this_century(),
                                              make_dataframe=True )

        self.submission_repo = create_autospec( ISubmissionRepo )

        self.obj = QuizGrader( self.work_repo, self.submission_repo )


    def test__grade_row_no_penalty( self ):
        sp = { 'analyze.return_value': 0 }
        for _ in range( 0, self.num ):
            m = MagicMock( **sp )
            self.penalizers.append( m )

        # fudge_points = 0
        # self.penalizer.get_fudge_points = MagicMock( return_value=fudge_points )

        row = self.work_repo.data.iloc[ 0 ]

        # call
        result = self.obj._grade_row( row )

        # check things called with expected values
        for gm in self.grade_methods:
            gm.grade.assert_called()
            for qid, column_name in self.work_repo.question_columns:
                # grade method called on each question field in row
                gm.grade.assert_any_call( row[ column_name ] )

        # check fudge points
        result_fudge_points = result[ 'data' ][ "quiz_submissions" ][ 0 ][ 'fudge_points' ]
        self.assertEqual( result_fudge_points, 0, "zero fudge points returned" )

        # check question scores
        result_question_dict = result[ 'data' ][ "quiz_submissions" ][ 0 ][ 'questions' ]
        for qid, cname in self.work_repo.question_columns:
            self.assertIn( qid, result_question_dict, "expected dictionary key" )
            self.assertIn( 'score',
                           result_question_dict[ qid ],
                           "expected internal dict key" )
            self.assertEqual( result_question_dict[ qid ][ 'score' ],
                              self.points_per_question,
                              "expected question score" )

    def test_grade_row_penalty( self ):
        total_possible = self.points_per_question * len(self.work_repo.question_columns)

        fudge_points = total_possible * -0.5

        sp = { 'analyze.return_value': -0.25 }
        for _ in range( 0, 2):
            self.penalizers.append( MagicMock( **sp ) )

        row = self.work_repo.data.iloc[ 0 ]

        # call
        result = self.obj._grade_row( row )

        # check things called with expected values
        for gm in self.grade_methods:
            gm.grade.assert_called()
            for qid, column_name in self.work_repo.question_columns:
                # grade method called on each question field in row
                gm.grade.assert_any_call( row[ column_name ] )

        # check fudge points
        result_fudge_points = result[ 'data' ][ "quiz_submissions" ][ 0 ][ 'fudge_points' ]
        self.assertEqual( result_fudge_points, fudge_points, "expected fudge points returned" )

        # check question scores
        result_question_dict = result[ 'data' ][ "quiz_submissions" ][ 0 ][ 'questions' ]
        for qid, cname in self.work_repo.question_columns:
            self.assertIn( qid, result_question_dict, "expected dictionary key" )
            self.assertIn( 'score',
                           result_question_dict[ qid ],
                           "expected internal dict key" )
            self.assertEqual( result_question_dict[ qid ][ 'score' ],
                              self.points_per_question,
                              "expected question score" )

        # check fudge points
        result_fudge_points = result[ 'data' ][ "quiz_submissions" ][ 0 ][ 'fudge_points' ]
        self.assertEqual( result_fudge_points, fudge_points, "expected fudge points returned" )

        # check question scores
        result_question_dict = result[ 'data' ][ "quiz_submissions" ][ 0 ][ 'questions' ]
        for qid, cname in self.work_repo.question_columns:
            self.assertIn( qid, result_question_dict, "expected dictionary key" )
            self.assertIn( 'score', result_question_dict[ qid ], "expected internal dict key" )
            self.assertEqual( result_question_dict[ qid ][ 'score' ], self.points_per_question,
                              "expected question score" )

    @patch('CanvasHacks.GradingHandlers.quiz_new.BlockedByOtherStudent')
    def test_grade_row_student_blocked( self , checkerMock):
        def a():
            raise StudentUnableToComplete
        checkerMock.analyze = MagicMock(side_effect=a)

        # Only care about being blocked if total score is 0
        [gm.pop() for gm in self.grade_methods]

        gp = { 'grade.return_value': 0 }
        for _ in range( 0, self.num ):
            self.grade_methods.append( MagicMock( **gp ) )


        # total_possible = self.points_per_question * len( self.work_repo.question_columns )
        #
        # fudge_points = total_possible * -0.5
        #
        # sp = { 'analyze.return_value': -0.25 }
        # for _ in range( 0, 2 ):
        #     self.penalizers.append( MagicMock( **sp ) )
        self.activity = create_autospec(BlockableActivity, grade_methods=self.grade_methods,
                                   penalizers=self.penalizers,
                                   corrections=self.corrections )

        self.assocRepo = MagicMock()

        self.obj = QuizGrader( self.work_repo, self.submission_repo, association_repo=self.assocRepo )

        row = self.work_repo.data.iloc[ 0 ]

        # call
        with self.assertRaises(StudentUnableToComplete):
            result = self.obj._grade_row( row )



    def test_grade( self ):
        self.skipTest( 'todo' )
        # todo Mixed credit and no credit students
        # todo Different total scores
        # todo Different attempts



"""
Created by adam on 3/12/20
"""
__author__ = 'adam'

from CanvasHacks.SkaaSteps.SendInitialWorkToReviewer import SendInitialWorkToReviewer
from CanvasHacks.SkaaSteps.SendMetareviewToReviewer import SendMetareviewToReviewer
from CanvasHacks.SkaaSteps.SendReviewToReviewee import SendReviewToReviewee

if __name__ == '__main__':
    pass

from ipywidgets import widgets
from IPython.display import display
from CanvasHacks import environment


def run_all_steps( SEND=True, download=True ):
    print( "====================== STEP 1 ======================" )
    step1 = SendInitialWorkToReviewer( course=environment.CONFIG.course, unit=environment.CONFIG.unit, send=SEND )
    step1.run( rest_timeout=5 )

    print( "\n====================== STEP 2 ======================" )
    step2 = SendReviewToReviewee( environment.CONFIG.course, environment.CONFIG.unit, send=SEND )
    step2.run( rest_timeout=5, download=download )

    print( "\n====================== STEP 3 ======================" )
    step3 = SendMetareviewToReviewer( environment.CONFIG.course, environment.CONFIG.unit, send=SEND )
    step3.run( rest_timeout=5, download=download )

    # Return in case need to check values on them
    return (step1, step2, step3)


def skaa_run_button( control_store, return_button=False, width='auto', **kwargs ):
    RUNNING = False

    def get_style( is_running=False ):
        return 'warning' if is_running else 'danger'

    def get_name( is_running=False ):
        return 'RUNNING' if is_running else 'RUN SKAA'

    # Create the button
    layout = widgets.Layout( width=width )
    b = widgets.Button( description=get_name( RUNNING ),
                        layout=layout,
                        button_style=get_style( RUNNING ) )

    def callback( change ):
        RUNNING = True
        b.description = get_name( RUNNING )
        b.button_style = get_style( RUNNING )
        #         print('beep')
        #         time.sleep(3)
        #         print('boop')

        steps = run_all_steps( SEND=True, download=True )

        RUNNING = False

        b.description = get_name( RUNNING )
        b.button_style = get_style( RUNNING )

        control_store['skaa_steps'] = steps

    b.on_click( callback )

    if return_button is True:
        # If putting inside a box or something else which will
        # call display
        return b
    else:
        display( b )



"""
Created by adam on 3/15/20
"""
__author__ = 'adam'

from abc import ABC

from CanvasHacks.Errors.grading import NonStringInContentField

if __name__ == '__main__':
    pass


class IGradingMethod( ABC ):
    """
    GradingMethods are in charge of taking whatever we are
    grading and determining the initial score based on the content.

    The grade method will return a float which the calling
    method should multiply against the available points.

    When multiple GradingMethods will be used to compute the
    initial score, each should be initialized with
    the pct_of_score containing a float such that the pct_of_score
    for all used GradingMethods add to 1

    Parent for those classes which return a
    positive float to indicate how much of the total
    possible credit a student should receive.

    Normally applied to questions
    """

    def grade( self, *args, **kwargs ):
        raise NotImplementedError



"""
Created by adam on 3/11/20
"""
__author__ = 'adam'

from unittest import TestCase

from CanvasHacks.Messaging.nagging import EssayNonSubmittersMessaging
from TestingBase import TestingBase
from factories.ModelFactories import student_factory
from factories.PeerReviewedFactories import unit_factory

from unittest.mock import patch, MagicMock

if __name__ == '__main__':
    pass


class TestEssayNonSubmittersMessaging( TestingBase ):
    def setUp(self):
        self.config_for_test()
        self.student = student_factory()
        self.unit = unit_factory()

    def test_prepare_message( self ):
        self.skipTest('todo')

    # @patch('CanvasHacks.Messaging.nagging.ConversationMessageSender')
    def test_send_message_to_student( self): #, messengerMock ):
        self.obj = EssayNonSubmittersMessaging( self.unit )
        self.obj.messenger = MagicMock()
        self.obj.messenger.send = MagicMock(return_value='taco')

        self.obj.send_message_to_student(self.student.student_id, self.student.first_name)

        self.obj.messenger.send.assert_called()

        # messengerMock.send.assert_called()

        # messengerMock.send.assert_called()
